from os.path import basename
from typing import List, Optional

from foundry.smb3parse.constants import BASE_OFFSET, PAGE_A000_ByTileset
from foundry.smb3parse.util.rom import Rom

WORLD_COUNT = 9  # includes warp zone

# W = WORLD_MAP
# OS = OFFSET

OS_SIZE = 2  # byte

TSA_OS_LIST = PAGE_A000_ByTileset
TSA_TABLE_SIZE = 0x400
TSA_TABLE_INTERVAL = TSA_TABLE_SIZE + 0x1C00


class ROM(Rom):
    MARKER_VALUE = bytes("SMB3FOUNDRY", "ascii")

    rom_data = bytearray()

    additional_data = ""

    path: str = ""
    name: str = ""

    W_INIT_OS_LIST: List[int] = []

    def __init__(self, path: Optional[str] = None):
        if not ROM.rom_data:
            if path is None:
                raise ValueError("Rom was not loaded!")

            ROM.load_from_file(path)

        super(ROM, self).__init__(ROM.rom_data)

        self.position = 0

    @staticmethod
    def get_tsa_data(object_set: int) -> bytearray:
        rom = ROM()

        tsa_index = rom.int(TSA_OS_LIST + object_set)

        if object_set == 0:
            # todo why is the tsa index in the wrong (seemingly) false?
            tsa_index += 1

        tsa_start = BASE_OFFSET + tsa_index * TSA_TABLE_INTERVAL

        return rom.read(tsa_start, TSA_TABLE_SIZE)

    @staticmethod
    def write_tsa_data(object_set: int, tsa_data: bytearray):
        rom = ROM()

        tsa_index = rom.int(TSA_OS_LIST + object_set)

        if object_set == 0:
            # todo why is the tsa index in the wrong (seemingly) false?
            tsa_index += 1

        tsa_start = BASE_OFFSET + tsa_index * TSA_TABLE_INTERVAL

        rom.bulk_write(tsa_data, tsa_start)

    @staticmethod
    def load_from_file(path: str):
        with open(path, "rb") as rom:
            data = bytearray(rom.read())

        ROM.path = path
        ROM.name = basename(path)

        additional_data_start = data.find(ROM.MARKER_VALUE)

        if additional_data_start == -1:
            ROM.rom_data = data
            ROM.additional_data = ""
        else:
            ROM.rom_data = data[:additional_data_start]

            additional_data_start += len(ROM.MARKER_VALUE)

            ROM.additional_data = data[additional_data_start:].decode("utf-8")

    @staticmethod
    def save_to_file(path: str, set_new_path=True):
        with open(path, "wb") as f:
            f.write(bytearray(ROM.rom_data))

        if ROM.additional_data:
            with open(path, "ab") as f:
                f.write(ROM.MARKER_VALUE)
                f.write(ROM.additional_data.encode("utf-8"))

        if set_new_path:
            ROM.path = path
            ROM.name = basename(path)

    @staticmethod
    def set_additional_data(additional_data):
        ROM.additional_data = additional_data

    @staticmethod
    def is_loaded() -> bool:
        return bool(ROM.path)

    def seek(self, position: int) -> int:
        if position > len(ROM.rom_data) or position < 0:
            return -1

        self.position = position

        return 0

    def get_byte(self, position: int = -1) -> int:
        if position >= 0:
            k = self.seek(position) >= 0
        else:
            k = self.position < len(ROM.rom_data)

        if k:
            return_byte = ROM.rom_data[self.position]
        else:
            return_byte = 0

        self.position += 1

        return return_byte

    def peek_byte(self, position: int = -1) -> int:
        old_position = self.position

        byte = self.get_byte(position)

        self.position = old_position

        return byte

    def bulk_read(self, count: int, position: int = -1) -> bytearray:
        if position >= 0:
            self.seek(position)
        else:
            position = self.position

        self.position += count

        return ROM.rom_data[position : position + count]

    def bulk_write(self, data: bytearray, position: int = -1):
        if position >= 0:
            self.seek(position)
        else:
            position = self.position

        self.position += len(data)

        ROM.rom_data[position : position + len(data)] = data



from typing import Optional

from foundry.smb3parse.levels import HEADER_LENGTH, LevelBase
from foundry.smb3parse.levels.level_header import LevelHeader
from foundry.smb3parse.objects.object_set import assert_valid_object_set_number
from foundry.smb3parse.util.rom import Rom


class Level(LevelBase):
    def __init__(self, rom: Rom, object_set_number: int, layout_address: int, enemy_address: int):
        super(Level, self).__init__(object_set_number, layout_address)

        self.object_set_number = object_set_number
        self.enemy_address = enemy_address

        self.world_map_position = None

        self._rom = rom

        self.header_address = self.layout_address - HEADER_LENGTH

        self.header_bytes = self._rom.read(self.header_address, HEADER_LENGTH)

        self.header = LevelHeader(self.header_bytes, self.object_set_number)

    def set_world_map_position(self, position):
        self.world_map_position = position

    def __eq__(self, other):
        if not isinstance(other, Level):
            return False

        return (
            self.object_set_number == other.object_set_number
            and self.layout_address == other.layout_address
            and self.enemy_address == other.enemy_address
        )

    @staticmethod
    def from_world_map(rom: Rom, world_map_position) -> Optional["Level"]:
        level_info = world_map_position.level_info

        if level_info is None:
            return None

        object_set_number, layout_address, enemy_address = level_info

        level = Level(rom, object_set_number, layout_address, enemy_address)

        level.set_world_map_position(world_map_position)

        return level

    @staticmethod
    def from_memory(rom: Rom, object_set_number: int, layout_address: int, enemy_address: int):
        assert_valid_object_set_number(object_set_number)

        level = Level(rom, object_set_number, layout_address, enemy_address)

        return level



from PySide6.QtWidgets import QFrame


# taken from https://stackoverflow.com/a/41068447/4252230
class HorizontalLine(QFrame):
    def __init__(self):
        super(HorizontalLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)



from functools import lru_cache

from PySide6.QtCore import QPoint
from PySide6.QtGui import QColor, QImage, QPainter, Qt

from foundry.game.File import ROM
from foundry.game.gfx.drawable import MASK_COLOR, apply_selection_overlay
from foundry.game.gfx.drawable.Tile import Tile
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import NESPalette, PaletteGroupProtocol


@lru_cache(2 ** 10)
def get_sprite(
    index: int,
    palette_group: PaletteGroupProtocol,
    palette_index: int,
    graphics_set: GraphicsSetProtocol,
    horizontal_mirror: bool = False,
    vertical_mirror: bool = False,
):
    if index > 0xFF:
        return Sprite(
            ROM().get_byte(index), palette_group, palette_index, graphics_set, horizontal_mirror, vertical_mirror
        )
    else:
        return Sprite(index, palette_group, palette_index, graphics_set, horizontal_mirror, vertical_mirror)


class Sprite:
    WIDTH: int = Tile.SIDE_LENGTH  # type: ignore
    HEIGHT: int = Tile.SIDE_LENGTH * 2  # type: ignore
    PIXEL_COUNT = WIDTH * HEIGHT

    _sprite_cache = {}

    def __init__(
        self,
        index: int,
        palette_group: PaletteGroupProtocol,
        palette_index: int,
        graphics_set: GraphicsSetProtocol,
        horizontal_mirror: bool = False,
        vertical_mirror: bool = False,
    ):
        self.index = index
        self.horizontal_mirror = horizontal_mirror
        self.vertical_mirror = vertical_mirror
        self.palette_group = palette_group
        self.palette_index = palette_index

        # can't hash list, so turn it into a string instead
        self._sprite_id = (index, str(palette_group), palette_index, graphics_set)

        self.top_tile = Tile(index, palette_group, palette_index, graphics_set)
        self.bottom_tile = Tile(index + 1, palette_group, palette_index, graphics_set)

        if vertical_mirror:
            self.top_tile, self.bottom_tile = self.bottom_tile, self.top_tile

        self.image = QImage(Sprite.WIDTH, Sprite.HEIGHT, QImage.Format_RGB888)

        painter = QPainter(self.image)
        painter.drawImage(QPoint(0, 0), self.top_tile.as_image().copy().mirrored(horizontal_mirror, vertical_mirror))
        painter.drawImage(
            QPoint(0, Tile.HEIGHT),  # type: ignore
            self.bottom_tile.as_image().copy().mirrored(horizontal_mirror, vertical_mirror),
        )
        painter.end()

    def draw(self, painter: QPainter, x, y, width, height, selected=False, transparent=False):
        sprite_attributes = (
            self._sprite_id,
            self.horizontal_mirror,
            self.vertical_mirror,
            width,
            height,
            selected,
            transparent,
        )

        if sprite_attributes not in Sprite._sprite_cache:
            image = self.image.copy()

            if width != Sprite.WIDTH or height != Sprite.HEIGHT:
                image = image.scaled(width, height)

            # mask out the transparent pixels first
            mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
            image.setAlphaChannel(mask)

            if not transparent:  # or self._whole_block_is_transparent:
                image = self._replace_transparent_with_background(image)

            if selected:
                apply_selection_overlay(image, mask)

            Sprite._sprite_cache[sprite_attributes] = image

        painter.drawImage(x, y, Sprite._sprite_cache[sprite_attributes])

    def _replace_transparent_with_background(self, image):
        # draw image on background layer, to fill transparent pixels
        background = image.copy()
        try:
            index = NESPalette[self.palette_group[self.palette_index][0]]
        except IndexError:
            return image
        background.fill(index)

        _painter = QPainter(background)
        _painter.drawImage(QPoint(), image)
        _painter.end()

        return background



from PySide6.QtCore import QRect, QSize, Qt
from PySide6.QtGui import QColor, QPainter, QPaintEvent
from PySide6.QtWidgets import QFormLayout, QLabel, QSizePolicy, QWidget

from foundry.gui.util import ease_color


class SizeBar(QWidget):
    def __init__(self, parent: QWidget, current_value: int, maximum_value: int):
        super(SizeBar, self).__init__(parent)

        self._maximum_value = maximum_value
        self._current_value = current_value
        self.value_color = QColor.black

    def sizeHint(self) -> QSize:
        size = super(SizeBar, self).sizeHint()
        size.setHeight(10)
        return size

    @property
    def current_value(self) -> int:
        return self._current_value

    @current_value.setter
    def current_value(self, value: int):
        self._current_value = value
        self.update()

    @property
    def maximum_value(self) -> int:
        return self._maximum_value

    @maximum_value.setter
    def maximum_value(self, value: int):
        self._maximum_value = value
        self.update()

    @property
    def display_color(self) -> QColor:
        if self.maximum_value == 0:
            return QColor(136, 8, 8)
        if self.current_value > self.maximum_value:
            overload = min((self.current_value - self.maximum_value) / 100, 1)
            return ease_color(QColor(136, 8, 8), QColor(255, 255, 0), overload)
        if 0.9 <= self.current_value / self.maximum_value <= 1:
            overload = ((self.current_value / self.maximum_value) - 0.9) * 10
            return ease_color(QColor(255, 255, 0), QColor(34, 139, 34), overload)
        else:
            return QColor(34, 139, 34)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        painter.fillRect(event.rect(), self.palette().base())

        total_length = max(self.current_value, self.maximum_value, 1)

        pixels_per_byte = event.rect().width() / total_length

        bar = QRect(event.rect())
        bar.setWidth(int(pixels_per_byte * self.current_value))

        painter.fillRect(bar, self.display_color)


class LevelSizeBar(QWidget):
    def __init__(self, parent, label: str, current_value: int, maximum_value: int):
        super(LevelSizeBar, self).__init__(parent)

        self.label = label
        self._current_value = current_value

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.setWhatsThis(
            "<b>Size Bar</b><br/>"
            "This bar shows, how much of the available space inside the level is currently taken up. It will turn "
            "red, when too many level objects have been placed (or if the level objects would result in more bytes, "
            "than the level originally had)."
        )

        self.bar = SizeBar(self, current_value, maximum_value)

        self.info_label = QLabel()
        self.info_label.setAlignment(Qt.AlignCenter)

        layout = QFormLayout(self)
        layout.addRow(self.info_label, self.bar)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

    def update(self):
        self.info_label.setText(f"{self.label}: {self.current_value}/{self.maximum_value} Bytes")
        return super(LevelSizeBar, self).update()

    @property
    def current_value(self) -> int:
        return self.bar.current_value

    @current_value.setter
    def current_value(self, value: int):
        self.bar.current_value = value
        self.update()

    @property
    def maximum_value(self) -> int:
        return self.bar.maximum_value

    @maximum_value.setter
    def maximum_value(self, value: int):
        self.bar.maximum_value = value
        self.update()



from typing import Optional

from attr import attrs
from PySide6.QtCore import QPoint, QRect, Signal, SignalInstance
from PySide6.QtGui import (
    QBrush,
    QCloseEvent,
    QColor,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
    Qt,
)
from PySide6.QtWidgets import QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.core.Position import Position
from foundry.game.gfx.drawable import MASK_COLOR
from foundry.game.gfx.drawable.Tile import Tile
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.Palette import NESPalette, PaletteGroup
from foundry.gui.CustomChildWindow import CustomChildWindow


@attrs(slots=True, auto_attribs=True)
class PatternViewerModel:
    graphics_set: GraphicsSet
    palette_group: PaletteGroup
    palette_index: int


class PatternViewerController(CustomChildWindow):
    graphics_set_changed: SignalInstance = Signal(GraphicsSet)  # type: ignore
    palette_group_changed: SignalInstance = Signal(PaletteGroup)  # type: ignore
    palette_index_changed: SignalInstance = Signal(int)  # type: ignore
    pattern_selected: SignalInstance = Signal(int)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self, parent: Optional[QWidget], graphics_set: GraphicsSet, palette_group: PaletteGroup, palette_index: int
    ):
        super().__init__(parent, "Pattern Viewer")

        self.model = PatternViewerModel(graphics_set, palette_group, palette_index)
        self.view = PatternViewerView(self, graphics_set, palette_group, palette_index)
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)

        self.view.pattern_selected.connect(self.pattern_selected.emit)

        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        def change_zoom(offset: int):
            self.view.zoom += offset

        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.addToolBar(self.toolbar)
        self.setStatusBar(QStatusBar(self))

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def graphics_set(self) -> GraphicsSet:
        return self.model.graphics_set

    @graphics_set.setter
    def graphics_set(self, value: GraphicsSet):
        self.model.graphics_set = value
        self.view.graphics_set = value
        self.graphics_set_changed.emit(self.graphics_set)
        self.view.update()

    @property
    def palette_group(self) -> PaletteGroup:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, value: PaletteGroup):
        self.model.palette_group = value
        self.view.palette_group = value
        self.palette_group_changed.emit(self.palette_group)
        self.view.update()

    @property
    def palette_index(self) -> int:
        return self.model.palette_index

    @palette_index.setter
    def palette_index(self, value: int):
        self.model.palette_index = value
        self.view.palette_index = value
        self.palette_index_changed.emit(self.palette_index)
        self.view.update()


class PatternViewerView(QWidget):
    pattern_selected: SignalInstance = Signal(int)  # type: ignore

    PATTERNS = 256
    PATTERNS_PER_ROW = 16
    PATTERNS_PER_COLUMN = 16

    def __init__(
        self,
        parent: Optional[QWidget],
        graphics_set: GraphicsSet,
        palette_group: PaletteGroup,
        palette_index: int,
        zoom: int = 4,
    ):
        super().__init__(parent)

        self.setMouseTracking(True)

        self.graphics_set = graphics_set
        self.palette_group = palette_group
        self.palette_index = palette_index
        self.zoom = zoom

    def resizeEvent(self, event: QResizeEvent):
        self.update()

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        self.setFixedSize(
            self.PATTERNS_PER_ROW * Tile.WIDTH * self.zoom,  # type: ignore
            self.PATTERNS_PER_COLUMN * Tile.HEIGHT * self.zoom,  # type: ignore
        )

    @property
    def pattern_scale(self) -> int:
        return Tile.WIDTH * self.zoom  # type: ignore

    def mouseMoveEvent(self, event: QMouseEvent):
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // self.pattern_scale, pos.y // self.pattern_scale

        dec_index = pos.y * self.PATTERNS_PER_ROW + pos.x
        hex_index = hex(dec_index).upper().replace("X", "x")

        status_message = f"Row: {pos.y}, Column: {pos.x}, Index: {dec_index} / {hex_index}"

        self.parent().statusBar().showMessage(status_message)  # type: ignore

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // self.pattern_scale, pos.y // self.pattern_scale

        index = pos.y * self.PATTERNS_PER_ROW + pos.x
        self.pattern_selected.emit(index)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        bg_color = NESPalette[self.palette_group[self.palette_index][0]]
        painter.setBrush(QBrush(bg_color))
        painter.drawRect(QRect(QPoint(0, 0), self.size()))

        for i in range(self.PATTERNS):
            tile = Tile(
                i, tuple(tuple(c for c in pal) for pal in self.palette_group), self.palette_index, self.graphics_set
            )

            x = (i % self.PATTERNS_PER_ROW) * self.pattern_scale
            y = (i // self.PATTERNS_PER_ROW) * self.pattern_scale

            image = tile.as_image(self.pattern_scale)
            mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
            image.setAlphaChannel(mask)
            painter.drawImage(x, y, image)



from foundry.game.ObjectDefinitions import (
    TilesetDefinition,
    get_object_metadata,
    object_set_to_definition,
)
from foundry.smb3parse.constants import TILESET_ENDINGS, TILESET_NAMES


class ObjectSet:
    def __init__(self, object_set_number: int):
        self.number = object_set_number

        self.name = TILESET_NAMES[self.number]

        self.definitions = get_object_metadata().__root__[object_set_to_definition[self.number]]

    def object_type(self, domain: int, index: int) -> int:
        domain_offset = domain * 0x1F

        if index <= 0x0F:
            return index + domain_offset
        else:
            return (index >> 4) + domain_offset + 16 - 1

    def get_definition_of(self, object_id: int) -> TilesetDefinition:
        return self.definitions.__root__[object_id]

    def get_ending_offset(self) -> int:
        return TILESET_ENDINGS[self.number]

    def get_object_byte_length(self, domain: int, object_id: int) -> int:
        definition = self.get_definition_of(self.object_type(domain, object_id))
        if definition.is_4byte:
            return 4
        else:
            return 3



from collections import deque
from typing import Generic, Optional, Protocol, TypeVar

T = TypeVar("T")


class UndoControllerProtocol(Protocol, Generic[T]):
    @property
    def state(self) -> T:
        ...

    def do(self, new_state: T) -> T:
        ...

    @property
    def can_undo(self) -> bool:
        ...

    def undo(self) -> T:
        ...

    @property
    def can_redo(self) -> bool:
        ...

    def redo(self) -> T:
        ...


class UndoController(Generic[T]):
    """
    A controller for handling both undo and redo

    Parameters
    ----------
    Generic : T
        The state being stored by the UndoController.
    """

    def __init__(self, initial_state: T, undo_stack: Optional[deque[T]] = None, redo_stack: Optional[deque[T]] = None):
        self._state: T = initial_state
        self.undo_stack: deque[T] = undo_stack if undo_stack is not None else deque()
        self.redo_stack: deque[T] = redo_stack if redo_stack is not None else deque()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.state}, {self.undo_stack}, {self.redo_stack})"

    @property
    def state(self) -> T:
        return self._state

    def do(self, new_state: T) -> T:
        """
        Does an action through the controller, adding it to the undo stack and clearing the redo
        stack, respectively.

        Parameters
        ----------
        new_state : T
            The new state to be stored.

        Returns
        -------
        T
            The new state that has been stored.
        """
        self.undo_stack.append(self.state)
        self.redo_stack = deque()
        self._state = new_state
        return self.state

    @property
    def can_undo(self) -> bool:
        """
        Determines if there is any states inside the undo stack.

        Returns
        -------
        bool
            If there is an undo state available.
        """
        return bool(len(self.undo_stack))

    def undo(self) -> T:
        """
        Undoes the last state, bring the previous.

        Returns
        -------
        T
            The new state that has been stored.
        """
        self.redo_stack.append(self.state)
        self._state = self.undo_stack.pop()
        return self.state

    @property
    def can_redo(self) -> bool:
        """
        Determines if there is any states inside the redo stack.

        Returns
        -------
        bool
            If there is an redo state available.
        """
        return bool(len(self.redo_stack))

    def redo(self) -> T:
        """
        Redoes the previously undone state.

        Returns
        -------
        T
            The new state that has been stored.
        """
        self.undo_stack.append(self.state)
        self._state = self.redo_stack.pop()
        return self.state



from typing import Optional

from PySide6.QtCore import QSize
from PySide6.QtGui import QPainter, QPaintEvent
from PySide6.QtWidgets import QWidget

from foundry.game.level.WorldMap import WorldMap


class WorldMapView(QWidget):
    def __init__(self, parent: Optional[QWidget], world: WorldMap):
        super(WorldMapView, self).__init__(parent)

        self.world = world
        self.zoom = 2

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        self.world.draw(painter, self.zoom)

    def sizeHint(self) -> QSize:
        return self.world.q_size * self.zoom



from json import loads

from foundry import tileset_data_path

with open(tileset_data_path, "r") as data:
    _tileset_data = loads(data.read())

TILESET_LEVEL_OFFSET = [int(tileset["location"], 16) for tileset in _tileset_data]
TILESET_NAMES: list[str] = [tileset["name"] for tileset in _tileset_data]
TILESET_RANGES = [range(int(tileset["start"], 16), int(tileset["end"], 16)) for tileset in _tileset_data]
TILESET_ENDINGS: list[int] = [tileset["ending"] for tileset in _tileset_data]

BASE_OFFSET = 0x10  # the size of the rom header identifying the rom

TILE_LEVEL_1 = 0x03
TILE_LEVEL_2 = 0x04
TILE_LEVEL_3 = 0x05
TILE_LEVEL_4 = 0x06
TILE_LEVEL_5 = 0x07
TILE_LEVEL_6 = 0x08
TILE_LEVEL_7 = 0x09
TILE_LEVEL_8 = 0x0A
TILE_LEVEL_9 = 0x0B
TILE_LEVEL_10 = 0x0C
TILE_MUSHROOM_HOUSE_1 = 0x50
TILE_MUSHROOM_HOUSE_2 = 0xE0
TILE_STAR_1 = 0x55
TILE_STAR_2 = 0xE9
TILE_SPIRAL_TOWER_1 = 0x5F
TILE_SPIRAL_TOWER_2 = 0xDF
TILE_DUNGEON_1 = 0x67
TILE_DUNGEON_2 = 0xEB
TILE_QUICKSAND = 0x68
TILE_PYRAMID = 0x69
TILE_PIPE = 0xBC
TILE_POND = 0xBF
TILE_CASTLE_BOTTOM = 0xC9
TILE_BOWSER_CASTLE = 0xCC  # TILE_BOWSERCASTLELL
TILE_HAND_TRAP = 0xE6
TILE_SPADE_HOUSE = 0xE8

OBJ_AUTOSCROLL = 0xD3

POWERUP_MUSHROOM = 0x01
POWERUP_FIREFLOWER = 0x02
POWERUP_RACCOON = 0x03
POWERUP_FROG = 0x04
POWERUP_TANOOKI = 0x05
POWERUP_HAMMER = 0x06
POWERUP_ADDITION_JUDGEMS = 0x07
POWERUP_ADDITION_PWING = 0x08
POWERUP_ADDITION_STARMAN = 0x09

PAGE_A000_ByTileset = 0x3C3F9
PAGE_C000_ByTileset = 0x3C3E6

Level_BG_Pages1 = 0x3D772
Level_BG_Pages2 = 0x3D789

Palette_By_Tileset = 0x377E2
PalSet_Maps = 0x36BE2

Level_TilesetIdx_ByTileset = 0x10010

AScroll_HorizontalInitMove = 0x1ECC
AScroll_MoveEndLoopSelect = 0x13944
AScroll_MovePlayer = 0x13CB0
AScroll_Movement = 0x13959
AScroll_MovementLoop = 0x13B42
AScroll_MovementLoopStart = 0x13B39
AScroll_MovementLoopTicks = 0x13B70
AScroll_MovementRepeat = 0x13A47
AScroll_VelAccel = 0x13B35

Title_DebugMenu = 0x30CD9
Title_PrepForWorldMap = 0x30CCB

Map_ByRowType = 0x193EC
Map_ByScrCol = 0x193FE
Map_ByXHi_InitIndex = 0x193DA
Map_Completable_Tiles = 0x18457
Map_EnterSpecialTiles = 0x14DBF
Map_LevelLayouts = 0x19422
Map_ObjSets = 0x19410
Map_Tile_Layouts = 0x185A8
Tile_Attributes_TS0 = 0x18410



from functools import cache
from typing import Type

from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.core.warnings.WarningType import WarningType


@cache
def type_to_pydantic_warning() -> dict[WarningType, Type[PydanticWarning]]:
    """
    Provide a dictionary to easily convert between a given warning type and pydantic warning

    Returns
    -------
    dict[WarningType, Type[PydanticWarning]]
        A dict containing a warning type and its corresponding pydantic warning.
    """
    from foundry.core.warnings.EnemyCompatibilityWarning import (
        PydanticEnemyCompatibilityWarning,
    )
    from foundry.core.warnings.ExtendToGroundWarning import (
        PydanticExtendToGroundWarning,
    )
    from foundry.core.warnings.InvalidObjectWarning import PydanticInvalidObjectWarning
    from foundry.core.warnings.InvalidPositionWarning import (
        PydanticInvalidPositionWarning,
    )
    from foundry.core.warnings.InvalidSizeWarning import PydanticInvalidSizeWarning
    from foundry.core.warnings.InvalidWarpWarning import PydanticInvalidWarpWarning
    from foundry.core.warnings.OutsideLevelBoundsWarning import (
        PydanticOutsideLevelBoundsWarning,
    )

    return {
        WarningType.enemy_compatibility: PydanticEnemyCompatibilityWarning,
        WarningType.invalid_extension_to_ground: PydanticExtendToGroundWarning,
        WarningType.invalid_type: PydanticInvalidObjectWarning,
        WarningType.invalid_position: PydanticInvalidPositionWarning,
        WarningType.invalid_size: PydanticInvalidSizeWarning,
        WarningType.invalid_warp: PydanticInvalidWarpWarning,
        WarningType.outside_level_bounds: PydanticOutsideLevelBoundsWarning,
    }


@cache
def type_to_warning() -> dict[WarningType, Type[Warning]]:
    """
    Provide a dictionary to easily convert between a given warning type and warning

    Returns
    -------
    dict[WarningType, Type[Warning]]
        A dict containing a warning type and its corresponding warning.
    """
    from foundry.core.warnings.EnemyCompatibilityWarning import (
        EnemyCompatibilityWarning,
    )
    from foundry.core.warnings.ExtendToGroundWarning import ExtendToGroundWarning
    from foundry.core.warnings.InvalidObjectWarning import InvalidObjectWarning
    from foundry.core.warnings.InvalidPositionWarning import InvalidPositionWarning
    from foundry.core.warnings.InvalidSizeWarning import InvalidSizeWarning
    from foundry.core.warnings.InvalidWarpWarning import InvalidWarpWarning
    from foundry.core.warnings.OutsideLevelBoundsWarning import (
        OutsideLevelBoundsWarning,
    )

    return {
        WarningType.enemy_compatibility: EnemyCompatibilityWarning,
        WarningType.invalid_extension_to_ground: ExtendToGroundWarning,
        WarningType.invalid_type: InvalidObjectWarning,
        WarningType.invalid_position: InvalidPositionWarning,
        WarningType.invalid_size: InvalidSizeWarning,
        WarningType.invalid_warp: InvalidWarpWarning,
        WarningType.outside_level_bounds: OutsideLevelBoundsWarning,
    }


def convert_pydantic_to_warning(warning: PydanticWarning) -> Type[Warning]:
    """
    Provides the class alternative for a given pydantic warning.

    Parameters
    ----------
    warning : PydanticWarning
        The pydantic warning to find the warning alternative of.

    Returns
    -------
    Type[Warning]
        The class that represents the pydantic warning.
    """
    return type_to_warning()[warning.type]



from typing import Union

from PySide6.QtCore import Qt, Signal, SignalInstance
from PySide6.QtGui import QIcon, QImage, QPixmap
from PySide6.QtWidgets import QApplication, QComboBox, QCompleter, QWidget

from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.EnemyItemFactory import EnemyItemFactory
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.LevelObjectFactory import LevelObjectFactory
from foundry.game.gfx.objects.ObjectLike import ObjectLike
from foundry.gui.ObjectIcon import get_minimal_icon_object
from foundry.smb3parse.objects import (
    MAX_DOMAIN,
    MAX_ENEMY_ITEM_ID,
    MAX_ID_VALUE,
    MIN_DOMAIN,
)


class ObjectDropdown(QComboBox):
    object_selected: SignalInstance = Signal(ObjectLike)

    def __init__(self, parent: QWidget):
        super(ObjectDropdown, self).__init__(parent)

        self.setEditable(True)
        self.setMaxVisibleItems(30)

        self.completer().setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self.completer().setFilterMode(Qt.MatchContains)

        self.setInsertPolicy(QComboBox.NoInsert)

        self.currentIndexChanged.connect(self._on_object_selected)

        # guard against overly long item descriptions
        self.setMaximumWidth(QApplication.primaryScreen().geometry().width() / 5)

        self.setWhatsThis(
            "<b>Object Dropdown</b><br/>"
            "Contains all objects and enemies/items, that can be placed in this type of level. Which are "
            "available depends on the object set, that is selected for this level.<br/>"
            "You can search, by typing in the name, or simply select it from the list. After selecting "
            "an object, you can place it by clicking the middle mouse button anywhere in the level."
        )

    def setFocus(self):
        super(ObjectDropdown, self).setFocus()

        self.lineEdit().selectAll()

    def set_object_set(self, object_set_index: int, graphic_set_index: int) -> None:
        factory = LevelObjectFactory(
            object_set_index, graphic_set_index, 0, [], vertical_level=False, size_minimal=True
        )

        self._on_object_factory_change(factory)

    def _on_object_selected(self, _):
        if self.currentIndex() == -1:
            return

        level_object = self.currentData(Qt.UserRole)

        self.object_selected.emit(level_object)

    def select_object(self, level_object: ObjectLike):
        """
        Called, when the current placeable object was selected from outside and we need to update the selection in the
        dropdown. This is not the selected object inside the level!

        :param level_object: The type of object, that was selected to be placeable in the level.
        """
        index_of_object = self.findText(level_object.name)

        if index_of_object == -1:
            raise LookupError(f"Couldn't find {level_object} in object dropdown.")

        was_blocked = self.blockSignals(True)
        self.setCurrentIndex(index_of_object)
        self.blockSignals(was_blocked)

    def _on_object_factory_change(self, object_factory: LevelObjectFactory) -> None:
        self.clear()

        self._object_factory = object_factory

        if self._object_factory is None:
            return

        # adds level objects
        for domain in range(MIN_DOMAIN, MAX_DOMAIN + 1):
            for static_object_id in range(0, 0x10):
                level_object = self._object_factory.from_properties(
                    domain, static_object_id, x=0, y=0, length=1, index=0
                )

                level_object = get_minimal_icon_object(level_object)

                self._add_item(level_object)

            for expanding_object_id in range(0x10, MAX_ID_VALUE, 0x10):
                level_object = self._object_factory.from_properties(
                    domain, expanding_object_id, x=0, y=0, length=1, index=0
                )

                level_object = get_minimal_icon_object(level_object)

                self._add_item(level_object)

        # insert visual separator between level objects and enemies/items
        self.insertSeparator(self.count())

        # adds enemies and items
        factory = EnemyItemFactory(object_factory.object_set, 0)

        for obj_index in range(MAX_ENEMY_ITEM_ID + 1):
            enemy_item = factory.from_properties(obj_index, x=0, y=0)

            self._add_item(enemy_item)

    def _add_item(self, level_object: Union[LevelObject, EnemyObject]):
        if not isinstance(level_object, (LevelObject, EnemyObject)):
            return

        if level_object.name in ["MSG_CRASH", "MSG_NOTHING", "MSG_POINTER"]:
            return

        icon = QIcon(QPixmap(self._resize_bitmap(level_object.as_image())))

        self.addItem(icon, level_object.name, level_object)

    @staticmethod
    def _resize_bitmap(source_image: QImage) -> QImage:
        image = source_image.scaled(Block.SIDE_LENGTH, Block.SIDE_LENGTH)

        return image



from typing import Protocol

from attr import attrs


class SizeProtocol(Protocol):
    """
    A two dimensional representation of a size of a given object.

    Attributes
    ----------
    width: int
        The width of the object being represented.
    height: int
        The height of the object being represented.
    """

    width: int
    height: int


@attrs(slots=True, auto_attribs=True)
class Size:
    """
    A two dimensional representation of a size, that uses ``attrs`` to create a basic
    implementation.

    Attributes
    ----------
    width: int
        The width of the object being represented.
    height: int
        The height of the object being represented.
    """

    width: int
    height: int



from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field
from PySide6.QtGui import QColor
from PySide6.QtWidgets import (
    QFormLayout,
    QHBoxLayout,
    QLayout,
    QMainWindow,
    QMenu,
    QPushButton,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from foundry import open_url
from foundry.gui.settings import SETTINGS

ID_PROP = "ID"


def clear_layout(layout):
    while layout.count():
        item = layout.takeAt(0)
        item.widget().deleteLater()


def ease_color(color1: QColor, color2: QColor, amount: float) -> QColor:
    """
    Eases between two colors, providing a combination of the two.

    Parameters
    ----------
    color1 : QColor
        The color which will be displayed with respect to amount.
    color2 : QColor
        The color which will be displayed with respect to 1 - amount.
    amount : float
        The amount as a decimal, where 1 represents everything and 0 represents nothing,
        that color1 will be shown and color2 will not.

    Returns
    -------
    QColor
        The intermediary color from color1 and color2.
    """
    alt_amount = 1 - amount
    return QColor(
        int(color1.red() * amount + color2.red() * alt_amount),
        int(color1.green() * amount + color2.green() * alt_amount),
        int(color1.blue() * amount + color2.blue() * alt_amount),
    )


class WhatIsThis(BaseModel):
    """
    A generic representation of the what's this description.  This enables a multi-line
    description to easily transfer from JSON and similar formats.

    Attributes
    ----------
    elements: list[str]
        A list of strings that represent each line of the description.
    """

    elements: list[str]

    @property
    def description(self) -> str:
        """
        The description of the `what's this` field in its entirety.

        Returns
        -------
        str
            The description of what the object is.
        """
        return "\n".join(self.elements)


class WidgetType(str, Enum):
    """
    A declaration of the widgets possible to be created through
    `JSON <https://en.wikipedia.org/wiki/JSON>`_ and
    `Pydantic <https://pydantic-docs.helpmanual.io/>`_.
    """

    button = "BUTTON"
    spinner = "SPINNER"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """
        A convenience method to quickly determine if a value is a valid enumeration.

        Parameters
        ----------
        value : str
            The value to check against the enumeration.

        Returns
        -------
        bool
            If the value is inside the enumeration.
        """
        return value in cls._value2member_map_


class Widget(BaseModel):
    """
    A generic representation of :class:`~PySide6.QtWidgets.QWidget`.

    Attributes
    ----------
    type: WidgetType
        The type of widget this widget represents.  This determines how constructors
        will treat the widget, often providing it additional parameters.
    parent_attribute_name: Optional[str]
        If provided, will be used to set this widget to its parent's name.
    what_is_this: Optional[WhatIsThis]
        The type hint for the given widget, if provided.
    """

    type: WidgetType
    parent_attribute_name: Optional[str]
    what_is_this: Optional[WhatIsThis]

    class Config:
        use_enum_values = True  # Allow storing the enum as a string


class Button(Widget):
    """
    A button representation of :class:`PySide6.QtWidgets.QPushButton`.

    Attributes
    ----------
    name: str
        The text that appears on the button.
    action: Optional[str]
        The name of the callable that will be acted upon when clicked.
    """

    name: str = Field(default_factory=Field(""))
    action: Optional[str]


class Spinner(Widget):
    """
    A spinner representation of :class:`PySide6.QtWidgets.QSpinBox`.


    Attributes
    ----------
    enabled: bool
        Decides if the spinner will be activated on start.
    minimum: Optional[int]
        Will provide a lower bound that will be applied to the spinner, if present.
    maximum: Optional[int]
        Will provide an upper bound that will be applied to the spinner, if present.
    hexadecimal: bool
        Decides if the spinner should use hex as its base.
    value_change_action: Optional[str]
        The name of the callable that will be acted upon when the value changes.
    """

    enabled: bool = Field(default_factor=Field(True))
    minimum: Optional[int]
    maximum: Optional[int]
    hexadecimal: bool = Field(default_factor=Field(False))
    value_change_action: Optional[str]


class WidgetCreator(BaseModel):
    """
    A generator for a :class:`~foundry.gui.util.Widget`.  Creates the widget dynamically from
    its type attribute to provide it additional information required to be a subclass of
    widget.

    For example a widget of type button will become a button, instead of a plain widget.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        return super().__init_subclass__(**kwargs)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def generate_widget(cls, v: dict) -> Widget:
        """
        The constructor for each specific widget.

        Parameters
        ----------
        v : dict
            The dictionary to create the widget.

        Returns
        -------
        Widget
            The created widget as defined by `v["type"]`.

        Raises
        ------
        NotImplementedError
            If the constructor does not have a valid constructor for `v["type"]`.
        """
        type_ = WidgetType(v["type"])
        if type_ == WidgetType.button:
            return Button(**v)
        if type_ == WidgetType.spinner:
            return Spinner(**v)
        raise NotImplementedError(f"There is no widget of type {type_}")

    @classmethod
    def validate(cls, v: dict) -> Widget:
        """
        Validates that the provided object is a valid Widget.

        Parameters
        ----------
        v : dict
            The dictionary to create the widget.

        Returns
        -------
        Widget
            If validated, a widget will be created in accordance to `generate_widget`.

        Raises
        ------
        TypeError
            If a dictionary is not provided.
        TypeError
            If the dictionary does not contain the key `"type"`.
        TypeError
            If the type provided is not inside :class:`~foundry.gui.util.WidgetType`.
        """
        if not isinstance(v, dict):
            raise TypeError("Dictionary required")
        if "type" not in v:
            raise TypeError("Must have a type")
        if not WidgetType.has_value(type_ := v["type"]):
            raise TypeError(f"{type_} is not a valid widget type")
        return cls.generate_widget(v)


class LayoutType(str, Enum):
    """
    A declaration of the layouts possible to be created through
    `JSON <https://en.wikipedia.org/wiki/JSON>`_ and
    `Pydantic <https://pydantic-docs.helpmanual.io/>`_.
    """

    horizontal = "HORIZONTAL"
    verticle = "VERTICLE"
    form = "FORM"

    @classmethod
    def has_value(cls, value):
        """
        A convenience method to quickly determine if a value is a valid enumeration.

        Parameters
        ----------
        value : str
            The value to check against the enumeration.

        Returns
        -------
        bool
            If the value is inside the enumeration.
        """

        return value in cls._value2member_map_


class LayoutMeta(BaseModel):
    """
    A generic representation of :class:`~PySide6.QtWidgets.QLayout`.

    Attributes
    ----------
    type: LayoutType
        The type of layout this layout represents.  This determines how constructors
        will treat the layout, often providing it additional parameters.
    """

    type: LayoutType

    class Config:
        use_enum_values = True  # Allow storing the enum as a string


class BoxLayout(LayoutMeta):
    """
    A layout which the widgets are laid in a linear fashion in a single direction.

    Attributes
    ----------
    widgets: list[WidgetCreator]
        A list of widgets that are inside the layout, which will start from the origin to the
        layout's ending point.
    """

    widgets: list[WidgetCreator]

    def get_widgets(self) -> list[Widget]:
        """
        A helper function to get the list of widgets with the correct typing hint.

        Returns
        -------
        list[Widget]
            The list of widgets as defined in `self.widgets`.
        """
        return self.widgets  # type: ignore


class Form(BaseModel):
    """
    A layout which sets a series of widgets adjacent to their respective labels.

    Attributes
    ----------
    label: str
        The description of the following widget to its side.
    widget: WidgetCreator
        The widget that is displayed in the given form.
    """

    label: str
    widget: WidgetCreator

    def get_widget(self) -> Widget:
        """
        A helper function to get the widget with the correct typing hint.

        Returns
        -------
        Widget
            The widget as defined in `self.widget`.
        """
        return self.widget  # type: ignore


class FormLayout(LayoutMeta):
    """
    A layout which widgets are adjacent to their respective labels.

    Attributes
    ----------
    forms: list[Form]
        The respective forms that compose the layout with their respective widgets and labels.
    """

    forms: list[Form]


class LayoutCreator(BaseModel):
    """
    A generator for a :class:`~foundry.gui.util.LayoutMeta`.  Creates the layout dynamically
    from its type attribute to provide it additional information required to be a subclass of
    layout.

    For example a layout of type horizontal will become a box layout, instead of a plain layout.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        return super().__init_subclass__(**kwargs)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def generate_layout(cls, v: dict) -> LayoutMeta:
        """
        The constructor for each specific layout.

        Parameters
        ----------
        v : dict
            The dictionary to create the layout.

        Returns
        -------
        LayoutMeta
            The created layout as defined by `v["type"]`

        Raises
        ------
        NotImplementedError
            If the constructor does not have a valid constructor for `v["type"]`.
        """

        type_ = LayoutType(v["type"])
        if type_ == LayoutType.horizontal or type_ == LayoutType.verticle:
            return BoxLayout(**v)
        if type_ == LayoutType.form:
            return FormLayout(**v)
        raise NotImplementedError(f"There is no layout of type {type_}")

    @classmethod
    def validate(cls, v):
        """
        Validates that the provided object is a valid LayoutMeta.

        Parameters
        ----------
        v : dict
            The dictionary to create the layout.

        Returns
        -------
        Widget
            If validated, a layout will be created in accordance to `generate_layout`.

        Raises
        ------
        TypeError
            If a dictionary is not provided.
        TypeError
            If the dictionary does not contain the key `"type"`.
        TypeError
            If the type provided is not inside :class:`~foundry.gui.util.LayoutType`.
        """
        if not isinstance(v, dict):
            raise TypeError("Dictionary required")
        if "type" not in v:
            raise TypeError("Must have a type")
        if not LayoutType.has_value(type_ := v["type"]):
            raise TypeError(f"{type_} is not a valid layout type")
        return cls.generate_layout(v)


class Layout(BaseModel):
    """
    A helper class to allow for easier use of :class:`~foundry.gui.util.LayoutCreator` as
    it does not provide type hints.

    Attributes
    ----------
    layout: LayoutCreator
        The layout creator, which generates the layout automatically in accordance to layout
        creator.
    """

    layout: LayoutCreator

    def get_layout(self) -> LayoutMeta:
        """
        A helper method that provides correct typing hints for layout.

        Returns
        -------
        LayoutMeta
            The LayoutMeta as described in `self.layout`.
        """
        return self.layout  # type: ignore


def create_widget(parent: QWidget, meta: Widget) -> QWidget:
    """
    Creates a widget from a :class:`~foundry.gui.util.Widget`, providing its actual
    implementation.

    Parameters
    ----------
    parent : QWidget
        The widget which houses the widget's actions.
    meta : Widget
        The instance that holds the widget's attributes.

    Returns
    -------
    QWidget
        The created instance of a QWidget in accordance to `meta`.

    Raises
    ------
    NotImplementedError
        If the constructor does not implement a valid implementation for a given `meta.type`.
    """
    if isinstance(meta, Button):
        widget = QPushButton(meta.name)
        if meta.action is not None:
            widget.clicked.connect(getattr(parent, meta.action))  # type: ignore
    elif isinstance(meta, Spinner):
        widget = QSpinBox()
        if not meta.enabled:
            widget.setEnabled(False)
        if meta.minimum is not None:
            widget.setMinimum(meta.minimum)
        if meta.maximum is not None:
            widget.setMaximum(meta.maximum)
        if meta.hexadecimal:
            widget.setDisplayIntegerBase(16)
            widget.setPrefix("0x")
        if meta.value_change_action is not None:
            widget.valueChanged.connect(getattr(parent, meta.value_change_action))  # type: ignore
    else:
        raise NotImplementedError(f"{meta.type} is not supported")

    if meta.parent_attribute_name is not None:
        setattr(parent, meta.parent_attribute_name, widget)
    if meta.what_is_this is not None:
        widget.setWhatsThis(meta.what_is_this.description)
    return widget


def setup_description(parent: QWidget, flags: dict):
    """
    Generates any descriptions for a widget.  Most commonly the `what's this` field.

    Parameters
    ----------
    parent : QWidget
        The widget to apply the description to.
    flags : dict
        The dict that describes the description.
    """
    if "what is this" in flags:
        parent.setWhatsThis(WhatIsThis(elements=flags["what is this"]).description)


def setup_layout(parent: QWidget, flags: dict) -> QLayout:
    """
    Creates a layout from a series of flags and a parent.  The layout will pass any events
    automatically to the parent and will work identically to one created through code.

    Parameters
    ----------
    parent : QWidget
        The parent of the layout.  Typically the layout becomes the main layout of the parent.
    flags : dict
        The dict that described the layout.

    Returns
    -------
    QLayout
        The instance of the well formed layout.

    Raises
    ------
    NotImplementedError
        If the provided `meta.type` is not implemened.
    """
    meta = Layout(**flags).get_layout()

    if isinstance(meta, BoxLayout):
        if meta.type == LayoutType.horizontal:
            layout = QHBoxLayout()
        elif meta.type == LayoutType.verticle:
            layout = QVBoxLayout()
        else:
            raise NotImplementedError(f"{meta} does not support layout type of {meta.type}")

        for layout_widget in meta.get_widgets():
            layout.addWidget(create_widget(parent, layout_widget))
    elif isinstance(meta, FormLayout):
        layout = QFormLayout()

        for form in meta.forms:
            layout.addRow(form.label, create_widget(parent, form.get_widget()))
    else:
        raise NotImplementedError(f"{meta.type} is not supported")

    parent.setLayout(layout)

    return layout


def setup_widget_menu(widget: QMainWindow, flags):
    for menu_name, menu in flags["menu"].items():
        qmenu = QMenu(menu["name"])
        if menu.get("attribute", False):
            setattr(widget, f"{menu_name}_menu", qmenu)
        if "action" in menu:
            qmenu.triggered.connect(getattr(widget, menu["action"]))
        menu_type = menu["type"]

        for index, option_group in enumerate(menu["options"]):
            if index != 0:
                qmenu.addSeparator()
            if menu_type == "actions":
                for base_name, option in option_group.items():
                    name: str = option["name"]
                    action = qmenu.addAction(name)
                    if option.get("attribute", False):
                        setattr(widget, f"{base_name}_action", action)
                    if "action" in option:
                        method = getattr(widget, option["action"])
                    elif "link" in option:

                        def call_link(url: str):
                            def call_link():
                                return open_url(url)

                            return call_link

                        method = call_link(option["link"])
                    else:
                        raise NotImplementedError
                    if option.get("wrapped", False):
                        action.triggered.connect(lambda *_, m=method: m())
                    else:
                        action.triggered.connect(method)

            elif menu_type == "settings":
                for option in option_group.values():
                    name: str = option["display_name"]
                    action = qmenu.addAction(name)
                    action.setProperty(ID_PROP, option["id"])
                    action.setCheckable(True)
                    action.setChecked(SETTINGS[option["name"]])
        widget.menuBar().addMenu(qmenu)


def setup_window(widget: QMainWindow, flags):
    setup_widget_menu(widget, flags)



from foundry.smb3parse.levels import DEFAULT_HORIZONTAL_HEIGHT
from foundry.smb3parse.objects import InLevelObject


class LevelObject(InLevelObject):
    def __init__(self, data: bytearray):
        super(LevelObject, self).__init__(data)

        if len(data) not in [3, 4]:
            raise ValueError(f"Length of the given data must be 3 or 4, was {len(data)}.")

        self.domain = data[0] >> 5
        self.y = data[0] & 0b0001_1111

        if self.y > DEFAULT_HORIZONTAL_HEIGHT:
            raise ValueError(
                f"Data designating y value cannot be higher than {DEFAULT_HORIZONTAL_HEIGHT}, was {self.y}."
            )

        self.id = data[1]
        self.x = data[2]

        if len(data) == 4:
            self.additional_length = data[3]



from typing import Optional

from PySide6.QtCore import QSize, Signal, SignalInstance
from PySide6.QtGui import Qt
from PySide6.QtWidgets import (
    QAbstractButton,
    QDialogButtonBox,
    QGridLayout,
    QVBoxLayout,
    QWidget,
)

from foundry.game.gfx.Palette import NESPalette
from foundry.gui.ColorButtonWidget import ColorButtonWidget
from foundry.gui.CustomDialog import CustomDialog


class ColorSelector(CustomDialog):
    ROWS = 4
    COLUMNS = 16

    ok_clicked: SignalInstance = Signal(int)  # type: ignore

    def __init__(self, parent: Optional[QWidget], title: str = "NES Color Table", size: Optional[QSize] = None):
        super().__init__(parent, title=title)

        self.size_ = size if not None else QSize(24, 24)

        self._selected_button = 0

        self.layout_ = QGridLayout()
        self.layout_.setSpacing(0)

        self._color_buttons = []
        for row in range(self.ROWS):
            for column in range(self.COLUMNS):
                color = NESPalette[row * self.COLUMNS + column]
                button = ColorButtonWidget(self, color, self.size_)
                button.setLineWidth(0)
                self._color_buttons.append(button)
                self.layout_.addWidget(button, row, column)
        self._color_buttons[0].selected = True

        for idx, btn in enumerate(self._color_buttons):
            btn.clicked.connect(lambda idx=idx: self._on_click(idx))

        self.buttons = QDialogButtonBox(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttons.clicked.connect(self._on_button)  # type: ignore

        layout = QVBoxLayout(self)
        layout.addLayout(self.layout_)

        layout.addWidget(self.buttons, alignment=Qt.AlignCenter)

    @property
    def last_selected_color_index(self) -> int:
        return self._selected_button

    def _on_click(self, index: int):
        self._color_buttons[self._selected_button].selected = False
        self._color_buttons[index].selected = True
        self._selected_button = index

    def _on_button(self, button: QAbstractButton):
        if button is self.buttons.button(QDialogButtonBox.Ok):  # ok button
            self.accept()
        else:
            self.reject()



from PySide6.QtWidgets import QSpinBox

SPINNER_MAX_VALUE = 0xFF_FF_FF  # arbitrary; 16,7 MB


class Spinner(QSpinBox):
    def __init__(self, parent, maximum=SPINNER_MAX_VALUE, base=16):
        super(Spinner, self).__init__(parent)

        self.setRange(0, maximum)
        self.setDisplayIntegerBase(base)

        if base == 16:
            self.setPrefix("0x")



from foundry.smb3parse.objects import InLevelObject


class EnemyItem(InLevelObject):
    def __init__(self, data):
        super(EnemyItem, self).__init__(data)

        if not len(data) == 3:
            raise ValueError(f"Length of the given data must be 3, was {len(data)}.")

        self.domain = 0

        self._id, self._x, self._y = data



from abc import ABC

from foundry.smb3parse.constants import (
    BASE_OFFSET,
    Level_TilesetIdx_ByTileset,
    Map_ByRowType,
    Map_ByScrCol,
    Map_ByXHi_InitIndex,
    Map_Completable_Tiles,
    Map_EnterSpecialTiles,
    Map_LevelLayouts,
    Map_ObjSets,
    Map_Tile_Layouts,
    PAGE_A000_ByTileset,
    PAGE_C000_ByTileset,
    Tile_Attributes_TS0,
)

OFFSET_SIZE = 2  # byte

ENEMY_BASE_OFFSET = BASE_OFFSET  # + 1
"""
One additional byte, at the beginning of every enemy data, where I don't know what does
"""

WORLD_MAP_BASE_OFFSET = BASE_OFFSET + 0xE000
"""
Offset for a lot of world related parsing.
"""

LEVEL_BASE_OFFSET = Level_TilesetIdx_ByTileset
"""
Offset for level related parsing. Currently only used in Header.
"""

LAYOUT_LIST_OFFSET = Map_Tile_Layouts

TILE_ATTRIBUTES_TS0_OFFSET = Tile_Attributes_TS0
"""
The first 4 bytes describe minimal indexes an overworld tile must have to be enterable.
"""

STRUCTURE_DATA_OFFSETS = Map_ByXHi_InitIndex
"""
This lists the start of a block of world meta data. 9 worlds means 9 times 2 bytes of offsets. The block starts with a
0x00, so that also marks the end of the block before it.
"""

LEVEL_Y_POS_LISTS = Map_ByRowType
"""
This list contains the offsets to the y positions/row indexes of the levels of a world map. Since world maps can have up
to 4 screens, the offset could points to 4 consecutive lists, so we need to know the amount of levels per screen, to
make sense of them.
"""

LEVEL_X_POS_LISTS = Map_ByScrCol
"""
This list contains the offsets to the x positions/column indexes of the levels in a world map. They are listed in a row
for all 4 screens.
"""

LEVEL_ENEMY_LIST_OFFSET = Map_ObjSets
"""
"""

LEVELS_IN_WORLD_LIST_OFFSET = Map_LevelLayouts
"""
The memory locations of levels inside a world map are listed in a row. This offset points to the memory locations of
these lists for every world. The first 2 bytes following this offset point to the levels in world 1, the next 2 for
world 2 etc.
"""

OFFSET_BY_OBJECT_SET_A000 = PAGE_A000_ByTileset
"""
A list of values, which specify which ROM page should be loaded into addresses 0xA000 - 0xBFFF for a given object set.
This is necessary, since the ROM is larger then the addressable RAM in the NES. The offsets of levels are always into
the RAM, which means, to address levels at different parts in the ROM these parts need to be loaded into the RAM first.
"""

OFFSET_BY_OBJECT_SET_C000 = PAGE_C000_ByTileset
"""
Same with the ROM page and addresses 0xC000 - 0xFFFF.
"""

WORLD_COUNT = 9  # includes warp zone

WORLD_MAP_HEIGHT = 9  # blocks
WORLD_MAP_SCREEN_WIDTH = 16  # blocks

FIRST_VALID_ROW = 2
"""
Tiles in rows before this one are part of the border and not valid overworld tiles.
"""

VALID_ROWS = range(FIRST_VALID_ROW, FIRST_VALID_ROW + WORLD_MAP_HEIGHT)
"""
A range of row values, where Mario could possibly stand.
"""

VALID_COLUMNS = range(WORLD_MAP_SCREEN_WIDTH)
"""
A range of column values, where Mario could possibly stand.
"""

COMPLETABLE_TILES_LIST = Map_Completable_Tiles
"""
A list of tile values, that are completable, like the Toad House.
"""

COMPLETABLE_LIST_END_MARKER = 0x00  # MCT_END
"""
A value, that specifies the end of the completable tiles, rather than a set address.
"""

SPECIAL_ENTERABLE_TILES_LIST = Map_EnterSpecialTiles
"""
A list of tile values, that are also enterable, like the castle and the toad house.
"""

SPECIAL_ENTERABLE_TILE_AMOUNT = 11  # the rom mistakenly uses 0x1A

WORLD_MAP_SCREEN_SIZE = WORLD_MAP_HEIGHT * WORLD_MAP_SCREEN_WIDTH  # bytes

# in bytes
HEADER_LENGTH = 9

# in blocks
LEVEL_MIN_LENGTH = 0x10
LEVEL_MAX_LENGTH = 0x100
LEVEL_LENGTH_INTERVAL = 0x10

DEFAULT_HORIZONTAL_HEIGHT = 27
DEFAULT_VERTICAL_WIDTH = 16


def is_valid_level_length(level_length: int) -> bool:
    return level_length in range(LEVEL_MIN_LENGTH, LEVEL_MAX_LENGTH + 1, LEVEL_LENGTH_INTERVAL)


class LevelBase(ABC):
    width: int
    height: int

    def __init__(self, object_set_number: int, layout_address: int):
        self.layout_address = layout_address

        self.object_set_number = object_set_number



WORLD_MAP_OBJECT_SET = 0x00
PLAINS_OBJECT_SET = 0x01
DUNGEON_OBJECT_SET = 0x02
HILLY_OBJECT_SET = 0x03
SKY_OBJECT_SET = 0x04
PIRANHA_PLANT_OBJECT_SET = 0x05
WATER_OBJECT_SET = 0x06
MUSHROOM_OBJECT_SET = 0x07
PIPE_OBJECT_SET = 0x08
DESERT_OBJECT_SET = 0x09
AIR_SHIP_OBJECT_SET = 0x0A
GIANT_OBJECT_SET = 0x0B
ICE_OBJECT_SET = 0x0C
CLOUDY_OBJECT_SET = 0x0D
UNDERGROUND_OBJECT_SET = 0x0E
SPADE_BONUS_OBJECT_SET = 0x0F

PLAINS_GRAPHICS_SET = 0x01
DUNGEON_GRAPHICS_SET = 0x02
HILLY_GRAPHICS_SET = 0x03
SKY_GRAPHICS_SET = 0x04
DESERT_GRAPHICS_SET = 0x09
CLOUDY_GRAPHICS_SET = 0x0D
UNDERGROUND_GRAPHICS_SET = 0x0E
ENEMY_ITEM_GRAPHICS_SET = 0x4C

MIN_OBJECT_SET = WORLD_MAP_OBJECT_SET
MAX_OBJECT_SET = 0x0F

# amount of consecutive objects in a group, that share the same byte length
OBJECT_GROUP_SIZE = 16


def assert_valid_object_set_number(object_set_number: int):
    if not is_valid_object_set_number(object_set_number):
        raise ValueError(f"Object set number {object_set_number} is invalid.")


def is_valid_object_set_number(object_set_number: int):
    return object_set_number in range(MIN_OBJECT_SET, MAX_OBJECT_SET + 1)



from typing import Optional

from PySide6.QtWidgets import (
    QComboBox,
    QDialogButtonBox,
    QFormLayout,
    QGroupBox,
    QVBoxLayout,
    QWidget,
)

from foundry.game.gfx.objects.Jump import Jump
from foundry.gui.CustomDialog import CustomDialog
from foundry.gui.Spinner import Spinner

JUMP_ACTIONS = {
    "Downward Pipe 1": 0,
    "Upward Pipe": 1,
    "Downward Pipe 2": 2,
    "Right Pipe": 3,
    "Left Pipe": 4,
    "Jump on Noteblock": 7,
    "Door": 8,
}


JUMP_INDEXES = [0, 1, 2, 3, 4, 0, 0, 5, 6, 0, 0, 0, 0, 0, 0, 0]


VERT_POSITIONS = ["00", "05", "08", "12", "16", "20", "23", "24"]

MAX_SCREEN_INDEX = 0x0F
MAX_HORIZ_POSITION = 0xFF


class JumpEditor(CustomDialog):
    def __init__(
        self, parent: Optional[QWidget], jump: Jump, is_horizontal: bool, suggested_max_size: int = MAX_SCREEN_INDEX
    ):
        super(JumpEditor, self).__init__(parent, "Jump Editor")

        self.jump = jump
        self.is_horizontal = is_horizontal

        self.screen_spinner = Spinner(parent=self, maximum=suggested_max_size, base=10)

        position_layout = QFormLayout()
        position_layout.addRow("Jump on screen:", self.screen_spinner)

        level_group_box = QGroupBox("Level position")
        level_group_box.setLayout(position_layout)

        self.exit_action = QComboBox(self)
        self.exit_action.addItems(JUMP_ACTIONS.keys())

        self.exit_horizontal = Spinner(parent=self, maximum=MAX_HORIZ_POSITION, base=10)

        self.exit_vertical = QComboBox(self)
        self.exit_vertical.addItems(VERT_POSITIONS)

        exit_layout = QFormLayout()
        exit_layout.addRow("Exit action:", self.exit_action)
        exit_layout.addRow("Exit position x:", self.exit_horizontal)
        exit_layout.addRow("Exit position y:", self.exit_vertical)

        exit_group_box = QGroupBox("Exit options")
        exit_group_box.setLayout(exit_layout)

        button_box = QDialogButtonBox()
        self.ok_button = button_box.addButton(QDialogButtonBox.Ok)
        self.ok_button.pressed.connect(self.on_ok)
        button_box.addButton(QDialogButtonBox.Cancel).pressed.connect(self.close)

        main_layout = QVBoxLayout()
        main_layout.addWidget(level_group_box)
        main_layout.addWidget(exit_group_box)
        main_layout.addWidget(button_box)

        self.setLayout(main_layout)

        self._set_widget_values()

    def _set_widget_values(self):
        self.screen_spinner.setValue(self.jump.screen_index)

        self.exit_action.setCurrentIndex(JUMP_INDEXES[self.jump.exit_action])
        self.exit_horizontal.setValue(self.jump.exit_horizontal)
        self.exit_vertical.setCurrentIndex(self.jump.exit_vertical & 0b111)

    @staticmethod
    def edit_jump(
        parent: Optional[QWidget], jump: Jump, is_horizontal: bool, suggested_max_size: int = MAX_SCREEN_INDEX
    ):
        jump_editor = JumpEditor(parent, jump, is_horizontal, suggested_max_size)

        jump_editor.exec_()

        return jump_editor.jump

    def on_ok(self):
        self.jump = Jump.from_properties(
            self.screen_spinner.value(),
            JUMP_ACTIONS[self.exit_action.currentText()],
            self.exit_horizontal.value(),
            self.exit_vertical.currentIndex() if self.is_horizontal else self.exit_vertical.currentIndex() + 8,
        )

        self.close()



from typing import List, Tuple
from warnings import warn

from PySide6.QtCore import QRect, QSize
from PySide6.QtGui import QImage, QPainter

from foundry.core.Position import Position, PositionProtocol
from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block, get_block
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.objects.GeneratorObject import GeneratorObject
from foundry.game.gfx.objects.ObjectLike import (
    EXPANDS_BOTH,
    EXPANDS_HORIZ,
    EXPANDS_NOT,
    EXPANDS_VERT,
)
from foundry.game.gfx.Palette import PaletteGroup
from foundry.game.ObjectDefinitions import EndType, GeneratorType, TilesetDefinition
from foundry.game.ObjectSet import ObjectSet
from foundry.smb3parse.objects.object_set import PLAINS_OBJECT_SET

SKY = 0
GROUND = 27

ENDING_STR = {
    EndType.UNIFORM: "Uniform",
    EndType.END_ON_TOP_OR_LEFT: "Top or Left",
    EndType.END_ON_BOTTOM_OR_RIGHT: "Bottom or Right",
    EndType.TWO_ENDS: "Top & Bottom/Left & Right",
}

ORIENTATION_TO_STR = {
    GeneratorType.HORIZONTAL: "Horizontal",
    GeneratorType.VERTICAL: "Vertical",
    GeneratorType.DIAG_DOWN_LEFT: "Diagonal ↙",
    GeneratorType.DESERT_PIPE_BOX: "Desert Pipe Box",
    GeneratorType.DIAG_DOWN_RIGHT: "Diagonal ↘",
    GeneratorType.DIAG_UP_RIGHT: "Diagonal ↗",
    GeneratorType.HORIZ_TO_GROUND: "Horizontal to the Ground",
    GeneratorType.HORIZONTAL_2: "Horizontal Alternative",
    GeneratorType.DIAG_WEIRD: "Diagonal Weird",  # up left?
    GeneratorType.SINGLE_BLOCK_OBJECT: "Single Block",
    GeneratorType.CENTERED: "Centered",
    GeneratorType.PYRAMID_TO_GROUND: "Pyramid to Ground",
    GeneratorType.PYRAMID_2: "Pyramid Alternative",
    GeneratorType.TO_THE_SKY: "To the Sky",
    GeneratorType.ENDING: "Ending",
}

# todo what is this, exactly?
ENDING_OBJECT_OFFSET = 0x1C8F9

# not all objects provide a block index for blank block
BLANK = -1

SCREEN_HEIGHT = 15
SCREEN_WIDTH = 16


class LevelObject(GeneratorObject):
    def __init__(
        self,
        data: bytearray,
        object_set: int,
        palette_group: PaletteGroup,
        graphics_set: GraphicsSetProtocol,
        objects_ref: List["LevelObject"],
        is_vertical: bool,
        index: int,
        size_minimal: bool = False,
    ):
        self.object_set = ObjectSet(object_set)

        self.graphics_set = graphics_set

        self.x_position = 0
        self.y_position = 0

        self.rendered_base_x = 0
        self.rendered_base_y = 0

        self.palette_group = palette_group

        self.index_in_level = index
        self.objects_ref = objects_ref
        self.vertical_level = is_vertical

        self.data = data

        self.selected = False

        self.size_minimal = size_minimal

        if self.size_minimal:
            self.ground_level = 0
        else:
            self.ground_level = GROUND

        self._length = 0
        self.secondary_length = 0

        self._setup()

    def _setup(self):
        data = self.data

        # where to look for the graphic data?
        self.domain = (data[0] & 0b1110_0000) >> 5

        # position relative to the start of the level (top)
        self.original_y = data[0] & 0b0001_1111
        self.y_position = self.original_y

        # position relative to the start of the level (left)
        self.original_x = data[1]
        self.x_position = self.original_x

        if self.vertical_level:
            offset = (self.x_position // SCREEN_WIDTH) * SCREEN_HEIGHT

            self.y_position += offset
            self.x_position %= SCREEN_WIDTH

        # describes what object it is
        self._obj_index = 0x00

        self.obj_index = data[2]

        object_data = self.object_set.get_definition_of(self.type)

        self.width = object_data.bmp_width
        self.height = object_data.bmp_height
        self.orientation = GeneratorType(object_data.orientation)
        self.ending = EndType(object_data.ending)
        self.name = object_data.description

        self.blocks = [int(block) for block in object_data.blocks]

        self.is_4byte = object_data.is_4byte

        if self.is_4byte and len(self.data) == 3:
            self.data.append(0)
        elif not self.is_4byte and len(data) == 4:
            del self.data[3]

        self._length = 0
        self.secondary_length = 0

        self._calculate_lengths()

        self.rect = QRect()

        self._render()

    @property
    def tsa_data(self) -> bytearray:
        return ROM.get_tsa_data(self.object_set.number)

    @property
    def definition(self) -> TilesetDefinition:
        return self.object_set.get_definition_of(self.type)

    @property
    def obj_index(self):
        return self._obj_index

    @obj_index.setter
    def obj_index(self, value):
        self._obj_index = value

        self.is_single_block = self.obj_index <= 0x0F

        domain_offset = self.domain * 0x1F

        if self.is_single_block:
            self.type = self.obj_index + domain_offset
        else:
            self.type = (self.obj_index >> 4) + domain_offset + 16 - 1

    @property
    def object_info(self):
        return self.object_set.number, self.domain, self.obj_index

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        if not self.is_4byte and not self.is_single_block:
            self._obj_index &= 0xF0
            self._obj_index |= value & 0x0F

        self._length = value

    def _calculate_lengths(self):
        if self.is_single_block:
            self._length = 1
        else:
            self._length = self.obj_index & 0b0000_1111

        if self.is_4byte:
            self.secondary_length = self.length
            self.length = self.data[3]

    def render(self):
        self._render()

    def _render(self):
        self.rendered_base_x = base_x = self.x_position
        self.rendered_base_y = base_y = self.y_position

        self.rendered_width = new_width = self.width
        self.rendered_height = new_height = self.height

        try:
            self.index_in_level = self.objects_ref.index(self)
        except ValueError:
            # the object has not been added yet, so stick with the one given in the constructor
            pass

        blocks_to_draw = []

        if self.orientation == GeneratorType.TO_THE_SKY:
            base_x = self.x_position
            base_y = SKY

            for _ in range(self.y_position):
                blocks_to_draw.extend(self.blocks[0 : self.width])

            blocks_to_draw.extend(self.blocks[-self.width :])

            new_height = self.y_position + (self.height - 1)

        elif self.orientation == GeneratorType.DESERT_PIPE_BOX:
            # segments are the horizontal sections, which are 8 blocks long
            # two of those are drawn per length bit
            # rows are the 4 block high rows Mario can walk in

            is_pipe_box_type_b = self.obj_index // 0x10 == 4

            rows_per_box = self.height
            lines_per_row = 4

            segment_width = self.width
            segments = (self.length + 1) * 2

            box_height = lines_per_row * rows_per_box

            new_width = segments * segment_width
            new_height = box_height

            for row_number in range(rows_per_box):
                for line in range(lines_per_row):
                    if is_pipe_box_type_b and row_number > 0 and line == 0:
                        # in pipebox type b we do not repeat the horizontal beams
                        line += 1

                    start = line * segment_width
                    stop = start + segment_width

                    for segment_number in range(segments):
                        blocks_to_draw.extend(self.blocks[start:stop])

            # draw another last row
            new_height += 1

            if is_pipe_box_type_b:
                # draw another open row
                start = segment_width
            else:
                # draw the first row again to close the box
                start = 0

            stop = start + segment_width

            for segment_number in range(segments):
                blocks_to_draw.extend(self.blocks[start:stop])

            # every line repeats the last block again for some reason
            for end_of_line in range(len(blocks_to_draw), 0, -new_width):
                blocks_to_draw.insert(end_of_line, blocks_to_draw[end_of_line - 1])

            new_width += 1

        elif self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            if self.ending == EndType.UNIFORM:
                new_height = (self.length + 1) * self.height
                new_width = (self.length + 1) * self.width

                left = [BLANK]
                right = [BLANK]
                slopes = self.blocks

            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                new_height = (self.length + 1) * self.height
                new_width = (self.length + 1) * (self.width - 1)  # without fill block

                if self.orientation in [GeneratorType.DIAG_DOWN_RIGHT, GeneratorType.DIAG_UP_RIGHT]:
                    fill_block = self.blocks[0:1]
                    slopes = self.blocks[1:]

                    left = fill_block
                    right = [BLANK]
                elif self.orientation == GeneratorType.DIAG_DOWN_LEFT:
                    fill_block = self.blocks[-1:]
                    slopes = self.blocks[0:-1]

                    right = fill_block
                    left = [BLANK]

                else:
                    fill_block = self.blocks[0:1]
                    slopes = self.blocks[1:]

                    right = [BLANK]
                    left = fill_block

            elif self.ending == EndType.END_ON_BOTTOM_OR_RIGHT:
                new_height = (self.length + 1) * self.height
                new_width = (self.length + 1) * (self.width - 1)  # without fill block

                fill_block = self.blocks[-1:]
                slopes = self.blocks[0:-1]

                left = [BLANK]
                right = fill_block
            else:
                # todo other two ends not used with diagonals?
                warn(f"{self.name} was not rendered.", RuntimeWarning)
                self.rendered_blocks = []
                return

            rows = []

            if self.height > self.width:
                slope_width = self.width
            else:
                slope_width = len(slopes)

            for y in range(new_height):
                amount_right = (y // self.height) * slope_width
                amount_left = new_width - slope_width - amount_right

                offset = y % self.height

                rows.append(amount_left * left + slopes[offset : offset + slope_width] + amount_right * right)

            if self.orientation in [GeneratorType.DIAG_UP_RIGHT]:
                for row in rows:
                    row.reverse()

            if self.orientation in [GeneratorType.DIAG_DOWN_RIGHT, GeneratorType.DIAG_UP_RIGHT]:
                if not self.height > self.width:
                    rows.reverse()

            if self.orientation == GeneratorType.DIAG_DOWN_RIGHT and self.height > self.width:
                # special case for 60 degree platform wire down right
                for row in rows:
                    row.reverse()

            if self.orientation in [GeneratorType.DIAG_UP_RIGHT]:
                base_y -= new_height - 1

            if self.orientation in [GeneratorType.DIAG_DOWN_LEFT]:
                base_x -= new_width - slope_width

            for row in rows:
                blocks_to_draw.extend(row)

        elif self.orientation in [GeneratorType.PYRAMID_TO_GROUND, GeneratorType.PYRAMID_2]:
            # since pyramids grow horizontally in both directions when extending
            # we need to check for new ground every time it grows

            base_x += 1  # set the new base_x to the tip of the pyramid

            for y in range(base_y, self.ground_level):
                new_height = y - base_y
                new_width = 2 * new_height

                bottom_row = QRect(base_x, y, new_width, 1)

                if any(
                    [
                        bottom_row.intersects(obj.get_rect()) and y == obj.get_rect().top()
                        for obj in self.objects_ref[0 : self.index_in_level]
                    ]
                ):
                    break

            base_x = base_x - (new_width // 2)

            blank = self.blocks[0]
            left_slope = self.blocks[1]
            left_fill = self.blocks[2]
            right_fill = self.blocks[3]
            right_slope = self.blocks[4]

            for y in range(new_height):
                blank_blocks = (new_width // 2) - (y + 1)
                middle_blocks = y  # times two

                blocks_to_draw.extend(blank_blocks * [blank])

                blocks_to_draw.append(left_slope)
                blocks_to_draw.extend(middle_blocks * [left_fill] + middle_blocks * [right_fill])
                blocks_to_draw.append(right_slope)

                blocks_to_draw.extend(blank_blocks * [blank])

        elif self.orientation == GeneratorType.ENDING:
            page_width = 16
            page_limit = page_width - self.x_position % page_width

            new_width = page_width + page_limit + 1
            new_height = (GROUND - 1) - SKY

            for y in range(SKY, GROUND - 1):
                blocks_to_draw.append(self.blocks[0])
                blocks_to_draw.extend([self.blocks[1]] * (new_width - 1))

            # todo magic number
            # ending graphics
            rom_offset = ENDING_OBJECT_OFFSET + self.object_set.get_ending_offset() * 0x60

            rom = ROM()

            ending_graphic_height = 6
            floor_height = 1

            y_offset = GROUND - floor_height - ending_graphic_height

            for y in range(ending_graphic_height):
                for x in range(page_width):
                    block_index = rom.get_byte(rom_offset + y * page_width + x - 1)

                    block_position = (y_offset + y) * new_width + x + page_limit + 1
                    blocks_to_draw[block_position] = block_index

            # the ending object is seemingly always 1 block too wide (going into the next screen)
            for end_of_line in range(len(blocks_to_draw) - 1, 0, -new_width):
                del blocks_to_draw[end_of_line]

            new_width -= 1

            # Mushroom/Fire flower/Star is categorized as an enemy

        elif self.orientation == GeneratorType.VERTICAL:
            new_height = self.length + 1
            new_width = self.width

            if self.ending == EndType.UNIFORM:
                if self.is_4byte:
                    # there is one VERTICAL 4-byte object: Vertically oriented X-blocks
                    # the width is the primary expansion
                    new_width = (self.obj_index & 0x0F) + 1

                for _ in range(new_height):
                    for y in range(self.height):
                        for x in range(new_width):
                            blocks_to_draw.append(self.blocks[y * self.height + x % self.width])

                # adjust height for giant blocks, so that the rect is correct
                new_height *= self.height

            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                # in case the drawn object is smaller than its actual size
                for y in range(min(self.height, new_height)):
                    offset = y * self.width
                    blocks_to_draw.extend(self.blocks[offset : offset + self.width])

                additional_rows = new_height - self.height

                # assume only the last row needs to repeat
                # todo true for giant blocks?
                if additional_rows > 0:
                    last_row = self.blocks[-self.width :]

                    for _ in range(additional_rows):
                        blocks_to_draw.extend(last_row)

            elif self.ending == EndType.END_ON_BOTTOM_OR_RIGHT:
                additional_rows = new_height - self.height

                # assume only the first row needs to repeat
                # todo true for giant blocks?
                if additional_rows > 0:
                    last_row = self.blocks[0 : self.width]

                    for _ in range(additional_rows):
                        blocks_to_draw.extend(last_row)

                # in case the drawn object is smaller than its actual size
                for y in range(min(self.height, new_height)):
                    offset = y * self.width
                    blocks_to_draw.extend(self.blocks[offset : offset + self.width])

            elif self.ending == EndType.TWO_ENDS:
                # object exists on ships
                top_row = self.blocks[0 : self.width]
                bottom_row = self.blocks[-self.width :]

                blocks_to_draw.extend(top_row)

                additional_rows = new_height - 2

                # repeat second to last row
                if additional_rows > 0:
                    for _ in range(additional_rows):
                        blocks_to_draw.extend(self.blocks[-2 * self.width : -self.width])

                if new_height > 1:
                    blocks_to_draw.extend(bottom_row)

        elif self.orientation in [GeneratorType.HORIZONTAL, GeneratorType.HORIZ_TO_GROUND, GeneratorType.HORIZONTAL_2]:
            new_width = self.length + 1

            downwards_extending_vine = 1, 0, 0x06
            wooden_sky_pole = 4, 0, 0x04

            if self.object_info in [downwards_extending_vine, wooden_sky_pole]:
                new_width -= 1

            if self.orientation == GeneratorType.HORIZ_TO_GROUND:
                # to the ground only, until it hits something
                for y in range(base_y, self.ground_level):
                    bottom_row = QRect(base_x, y, new_width, 1)

                    if any(
                        [
                            bottom_row.intersects(obj.get_rect()) and y == obj.get_rect().top()
                            for obj in self.objects_ref[0 : self.index_in_level]
                        ]
                    ):
                        new_height = y - base_y
                        break
                else:
                    # nothing underneath this object, extend to the ground
                    new_height = self.ground_level - base_y

                if self.is_single_block:
                    new_width = self.length

                min_height = min(self.height, 2)

                new_height = max(min_height, new_height)

            elif self.orientation == GeneratorType.HORIZONTAL_2 and self.ending == EndType.TWO_ENDS:
                # floating platforms seem to just be one shorter for some reason
                new_width -= 1
            else:
                new_height = self.height + self.secondary_length

            if self.ending == EndType.UNIFORM and not self.is_4byte:
                for y in range(new_height):
                    offset = (y % self.height) * self.width

                    for _ in range(0, new_width):
                        blocks_to_draw.extend(self.blocks[offset : offset + self.width])

                # in case of giant blocks
                new_width *= self.width

            elif self.ending == EndType.UNIFORM and self.is_4byte:
                # 4 byte objects
                top = self.blocks[0:1]
                bottom = self.blocks[-1:]

                new_height = self.height + self.secondary_length

                # ceilings are one shorter than normal
                if self.height > self.width:
                    new_height -= 1

                if self.orientation == GeneratorType.HORIZONTAL_2:
                    for _ in range(0, new_height - 1):
                        blocks_to_draw.extend(new_width * top)

                    blocks_to_draw.extend(new_width * bottom)
                else:
                    blocks_to_draw.extend(new_width * top)

                    for _ in range(1, new_height):
                        blocks_to_draw.extend(new_width * bottom)

            elif self.ending == EndType.END_ON_TOP_OR_LEFT:
                for y in range(new_height):
                    offset = y * self.width

                    blocks_to_draw.append(self.blocks[offset])

                    for x in range(1, new_width):
                        blocks_to_draw.append(self.blocks[offset + 1])

            elif self.ending == EndType.END_ON_BOTTOM_OR_RIGHT:
                for y in range(new_height):
                    offset = y * self.width

                    for x in range(new_width - 1):
                        blocks_to_draw.append(self.blocks[offset])

                    blocks_to_draw.append(self.blocks[offset + self.width - 1])

            elif self.ending == EndType.TWO_ENDS:
                if self.orientation == GeneratorType.HORIZONTAL and self.is_4byte:
                    # flat ground objects have an artificial limit of 2 lines
                    if (
                        self.object_set.number == PLAINS_OBJECT_SET
                        and self.domain == 0
                        and self.obj_index in range(0xC0, 0xE0)
                    ):
                        self.height = new_height = min(2, self.secondary_length + 1)
                    else:
                        new_height = self.secondary_length + 1

                if self.width > len(self.blocks):
                    raise ValueError(f"{self} does not provide enough blocks to fill a row.")
                else:
                    start = 0
                    end = self.width

                for y in range(self.height):
                    new_start = y * self.width
                    new_end = (y + 1) * self.width

                    if new_end > len(self.blocks):
                        # repeat the last line of blocks to fill the object
                        pass
                    else:
                        start = new_start
                        end = new_end

                    left, *middle, right = self.blocks[start:end]

                    blocks_to_draw.append(left)
                    blocks_to_draw.extend(middle * (new_width - 2))
                    blocks_to_draw.append(right)

                if not len(blocks_to_draw) % self.height == 0:
                    warn(f"Blocks to draw are not divisible by height. {self}", RuntimeWarning)

                new_width = int(len(blocks_to_draw) / self.height)

                top_row = blocks_to_draw[0:new_width]
                middle_blocks = blocks_to_draw[new_width : new_width * 2]
                bottom_row = blocks_to_draw[-new_width:]

                blocks_to_draw = top_row

                for y in range(1, new_height - 1):
                    blocks_to_draw.extend(middle_blocks)

                if new_height > 1:
                    blocks_to_draw.extend(bottom_row)
        else:
            if not self.orientation == GeneratorType.SINGLE_BLOCK_OBJECT:
                warn(f"Didn't render {self.name}", RuntimeWarning)
                # breakpoint()

            if self.name.lower() == "black boss room background":
                new_width = SCREEN_WIDTH
                new_height = SCREEN_HEIGHT

                base_x = self.x_position // SCREEN_WIDTH * SCREEN_WIDTH
                base_y = 0

                blocks_to_draw = SCREEN_WIDTH * SCREEN_HEIGHT * [self.blocks[0]]

        # for not yet implemented objects and single block objects
        if blocks_to_draw:
            self.rendered_blocks = blocks_to_draw
        else:
            self.rendered_blocks = self.blocks

        self.rendered_width = new_width
        self.rendered_height = new_height
        self.rendered_base_x = base_x
        self.rendered_base_y = base_y

        if new_width and not self.rendered_height == len(self.rendered_blocks) / new_width:
            warn(
                f"Not enough Blocks for calculated height: {self.name}. "
                f"Blocks for height: {len(self.rendered_blocks) / new_width}. Rendered height: {self.rendered_height}",
                RuntimeWarning,
            )

            self.rendered_height = len(self.rendered_blocks) / new_width
        elif new_width == 0:
            warn(
                f"Calculated Width is 0, setting to 1: {self.name}. "
                f"Blocks to draw: {len(self.rendered_blocks)}. Rendered height: {self.rendered_height}",
                RuntimeWarning,
            )

            self.rendered_width = 1

        self.rect = QRect(self.rendered_base_x, self.rendered_base_y, self.rendered_width, self.rendered_height)

    def draw(self, painter: QPainter, block_length, transparent):
        for index, block_index in enumerate(self.rendered_blocks):
            if block_index == BLANK:
                continue

            x = self.rendered_base_x + index % self.rendered_width
            y = self.rendered_base_y + index // self.rendered_width

            self._draw_block(painter, block_index, x, y, block_length, transparent)

    def _draw_block(self, painter: QPainter, block_index, x, y, block_length, transparent):
        block = get_block(
            block_index,
            tuple(tuple(c for c in pal) for pal in self.palette_group),
            self.graphics_set,
            bytes(self.tsa_data),
        )

        block.draw(
            painter,
            x * block_length,
            y * block_length,
            block_length=block_length,
            selected=self.selected,
            transparent=transparent,
        )

    def move_by(self, dx: int, dy: int):
        new_x = self.rendered_base_x + dx
        new_y = self.rendered_base_y + dy

        self.position = Position(new_x, new_y)

    @property
    def position(self) -> PositionProtocol:
        return Position(self.x_position, self.y_position)

    @position.setter
    def position(self, position: PositionProtocol) -> None:
        x, y = position.x, position.y

        # todo also check for the upper bounds
        x = max(0, x)

        if self.orientation == GeneratorType.TO_THE_SKY:
            y = self.rendered_base_y + y
        else:
            y = max(0, y)

        x_diff = self.x_position - self.rendered_base_x
        y_diff = self.y_position - self.rendered_base_y

        self.rendered_base_x = int(x)
        self.rendered_base_y = int(y)

        self.x_position = self.rendered_base_x + x_diff
        self.y_position = self.rendered_base_y + y_diff

        self._render()

    def expands(self):
        expands = EXPANDS_NOT

        if self.is_single_block:
            return expands

        if self.is_4byte:
            expands |= EXPANDS_BOTH

        elif self.orientation in [
            GeneratorType.HORIZONTAL,
            GeneratorType.HORIZONTAL_2,
            GeneratorType.HORIZ_TO_GROUND,
        ] or self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            expands |= EXPANDS_HORIZ

        elif self.orientation in [GeneratorType.VERTICAL, GeneratorType.DIAG_WEIRD]:
            expands |= EXPANDS_VERT

        return expands

    def primary_expansion(self):
        if self.orientation in [
            GeneratorType.HORIZONTAL,
            GeneratorType.HORIZONTAL_2,
            GeneratorType.HORIZ_TO_GROUND,
        ] or self.orientation in [
            GeneratorType.DIAG_DOWN_LEFT,
            GeneratorType.DIAG_DOWN_RIGHT,
            GeneratorType.DIAG_UP_RIGHT,
            GeneratorType.DIAG_WEIRD,
        ]:
            if self.is_4byte:
                return EXPANDS_VERT
            else:
                return EXPANDS_HORIZ
        elif self.orientation in [GeneratorType.VERTICAL]:
            if self.is_4byte:
                return EXPANDS_HORIZ
            else:
                return EXPANDS_VERT
        else:
            return EXPANDS_BOTH

    def resize_x(self, x: int):
        if self.expands() & EXPANDS_HORIZ == 0:
            return

        if self.primary_expansion() == EXPANDS_HORIZ:
            length = x - self.x_position

            length = max(0, length)
            length = min(length, 0x0F)

            base_index = (self.obj_index // 0x10) * 0x10

            self.obj_index = base_index + length
            self.data[2] = self.obj_index
        else:
            length = x - self.x_position
            length = max(0, length)
            length = min(length, 0xFF)

            if self.is_4byte:
                self.data[3] = length
            else:
                raise ValueError("Resize impossible", self)

        self._calculate_lengths()

        self._render()

    def resize_y(self, y: int):
        if self.expands() & EXPANDS_VERT == 0:
            return

        if self.primary_expansion() == EXPANDS_VERT:
            length = y - self.y_position

            length = max(0, length)
            length = min(length, 0x0F)

            base_index = (self.obj_index // 0x10) * 0x10

            self.obj_index = base_index + length
            self.data[2] = self.obj_index
        else:
            length = y - self.y_position
            length = max(0, length)
            length = min(length, 0xFF)

            if self.is_4byte:
                self.data[3] = length
            else:
                raise ValueError("Resize impossible", self)

        self._calculate_lengths()

        self._render()

    def resize_by(self, dx: int, dy: int):
        if dx:
            self.resize_x(self.x_position + dx)

        if dy:
            self.resize_y(self.y_position + dy)

    def increment_type(self):
        self.change_type(True)

    def decrement_type(self):
        self.change_type(False)

    def change_type(self, increment: bool):
        value = 1 if increment else -1

        new_type = self.obj_index + value

        if new_type < 0 and self.domain > 0:
            new_domain = self.domain - 1
            new_type = 0xF0
        elif new_type > 0xFF and self.domain < 7:
            new_domain = self.domain + 1
            new_type = 0x00
        else:
            new_type = min(0xFF, new_type)
            new_type = max(0, new_type)

            new_domain = self.domain

        self.data[0] &= 0b0001_1111
        self.data[0] |= new_domain << 5

        self.data[2] = new_type

        self._setup()

    def __contains__(self, item: Tuple[int, int]) -> bool:
        x, y = item

        return self.point_in(x, y)

    def point_in(self, x: int, y: int) -> bool:
        return self.rect.contains(x, y)

    def get_status_info(self) -> List[tuple]:
        return [
            ("x", self.rendered_base_x),
            ("y", self.rendered_base_y),
            ("Width", self.rendered_width),
            ("Height", self.rendered_height),
            ("Orientation", ORIENTATION_TO_STR[self.orientation]),
            ("Ending", ENDING_STR[self.ending]),
        ]

    def display_size(self, zoom_factor: int = 1):
        return QSize(self.rendered_width * Block.SIDE_LENGTH, self.rendered_height * Block.SIDE_LENGTH) * zoom_factor

    def as_image(self) -> QImage:
        self.rendered_base_x = 0
        self.rendered_base_y = 0

        image = QImage(
            QSize(self.rendered_width * Block.SIDE_LENGTH, self.rendered_height * Block.SIDE_LENGTH),
            QImage.Format_RGB888,
        )

        bg_color = self.palette_group.background_color

        image.fill(bg_color)

        painter = QPainter(image)

        self.draw(painter, Block.SIDE_LENGTH, True)

        return image

    def to_bytes(self) -> bytearray:
        data = bytearray()

        if self.vertical_level:
            # todo from vertical to non-vertical is bugged, because it
            # seems like you can't convert the coordinates 1:1
            # there seems to be ambiguity

            offset = self.y_position // SCREEN_HEIGHT

            x_position = self.x_position + offset * SCREEN_WIDTH
            y_position = self.y_position % SCREEN_HEIGHT
        else:
            x_position = self.x_position
            y_position = self.y_position

        if self.orientation in [GeneratorType.PYRAMID_TO_GROUND, GeneratorType.PYRAMID_2]:
            x_position = self.rendered_base_x - 1 + self.rendered_width // 2

        data.append((self.domain << 5) | y_position)
        data.append(x_position)

        if not self.is_4byte and not self.is_single_block:
            third_byte = (self.obj_index & 0xF0) + self.length
        else:
            third_byte = self.obj_index

        data.append(third_byte)

        if self.is_4byte:
            data.append(self.length)

        return data

    def __repr__(self) -> str:
        return f"LevelObject {self.name} at {self.x_position}, {self.y_position}"

    def __eq__(self, other):
        if not isinstance(other, LevelObject):
            return False
        else:
            return self.to_bytes() == other.to_bytes()

    def __lt__(self, other):
        return self.index_in_level < other.index_in_level



from enum import Enum
from functools import cache
from json import loads

from pydantic import BaseModel

from foundry import tileset_definitions
from foundry.core.warnings.OutsideLevelBoundsWarning import OutsideLevelBoundsWarning
from foundry.core.warnings.Warning import Warning
from foundry.game.Definitions import Definition
from foundry.smb3parse.objects.object_set import (
    AIR_SHIP_OBJECT_SET,
    CLOUDY_OBJECT_SET,
    DESERT_OBJECT_SET,
    DUNGEON_OBJECT_SET,
    GIANT_OBJECT_SET,
    HILLY_OBJECT_SET,
    ICE_OBJECT_SET,
    MUSHROOM_OBJECT_SET,
    PIPE_OBJECT_SET,
    PIRANHA_PLANT_OBJECT_SET,
    PLAINS_OBJECT_SET,
    SKY_OBJECT_SET,
    SPADE_BONUS_OBJECT_SET,
    UNDERGROUND_OBJECT_SET,
    WATER_OBJECT_SET,
    WORLD_MAP_OBJECT_SET,
)


class GeneratorType(int, Enum):
    """
    Level objects are generated using different methods, depending on their generator type. Some objects extend until
    they hit another object, some extend up to the sky. To identify in what way a specific type of level object is
    constructed, this enum lists the known generator types.
    """

    HORIZONTAL = 0
    VERTICAL = 1  # vertical downward
    DIAG_DOWN_LEFT = 2
    DESERT_PIPE_BOX = 3
    DIAG_DOWN_RIGHT = 4
    DIAG_UP_RIGHT = 5
    HORIZ_TO_GROUND = 6
    HORIZONTAL_2 = 7  # special case of horizontal, floating boxes, ceilings
    DIAG_WEIRD = 8  #
    SINGLE_BLOCK_OBJECT = 9
    CENTERED = 10  # like spinning platforms
    PYRAMID_TO_GROUND = 11  # to the ground or next object
    PYRAMID_2 = 12  # doesn't exist?
    TO_THE_SKY = 13
    ENDING = 14


class EndType(int, Enum):
    """
    Some level objects have blocks designated to be used at their ends. For example pipes, which can be extended, but
    always end at one side with the same couple of blocks. To keep track of where those special blocks are to be placed,
    this enum is used. When the value is TWO_ENDS they are always on opposite sides and whether they are left and right
    or top and bottom depends on the generator type of the object.
    """

    UNIFORM = 0
    END_ON_TOP_OR_LEFT = 1
    END_ON_BOTTOM_OR_RIGHT = 2
    TWO_ENDS = 3


class TilesetDefinition(Definition):
    domain: int
    min_value: int
    max_value: int
    bmp_width: int
    bmp_height: int
    blocks: list[int]
    orientation: GeneratorType
    ending: EndType
    size: int = 3
    check_level_bounds: bool = True

    @property
    def object_design_length(self) -> int:
        return len(self.blocks)

    @property
    def is_4byte(self) -> bool:
        return self.size == 4

    def get_warnings(self) -> list[Warning]:
        warnings: list[Warning] = [warning.to_warning() for warning in self.warnings.copy()]  # type: ignore
        if self.check_level_bounds:
            warnings.append(OutsideLevelBoundsWarning())
        return warnings


class Tileset(BaseModel):
    __root__: list[TilesetDefinition]


class Tilesets(BaseModel):
    __root__: list[Tileset]


@cache
def get_object_metadata() -> Tilesets:
    with open(tileset_definitions, "r") as f:
        return Tilesets(__root__=loads(f.read()))


object_set_to_definition = {
    WORLD_MAP_OBJECT_SET: 0,
    PLAINS_OBJECT_SET: 1,
    MUSHROOM_OBJECT_SET: 1,
    SPADE_BONUS_OBJECT_SET: 1,
    HILLY_OBJECT_SET: 2,
    SKY_OBJECT_SET: 3,
    DUNGEON_OBJECT_SET: 4,
    AIR_SHIP_OBJECT_SET: 5,
    CLOUDY_OBJECT_SET: 6,
    DESERT_OBJECT_SET: 7,
    WATER_OBJECT_SET: 8,
    PIPE_OBJECT_SET: 8,
    PIRANHA_PLANT_OBJECT_SET: 9,
    GIANT_OBJECT_SET: 9,
    ICE_OBJECT_SET: 10,
    UNDERGROUND_OBJECT_SET: 11,
}



import abc

from foundry.smb3parse.levels import LevelBase


class LevelLike(LevelBase, abc.ABC):
    width: int
    height: int

    def __init__(self, object_set_number, layout_address):
        super(LevelLike, self).__init__(object_set_number, layout_address)

        self.changed = False

    @abc.abstractmethod
    def index_of(self, obj):
        pass

    @abc.abstractmethod
    def object_at(self, x, y):
        pass

    @abc.abstractmethod
    def get_object_names(self):
        pass

    @abc.abstractmethod
    def get_all_objects(self):
        pass

    @abc.abstractmethod
    def get_object(self, index):
        pass

    @abc.abstractmethod
    def remove_object(self, obj):
        pass

    @abc.abstractmethod
    def draw(self, dc, block_length, transparency, show_expansion):
        pass

    @abc.abstractmethod
    def to_bytes(self):
        pass

    @abc.abstractmethod
    def from_bytes(self, object_data, enemy_data):
        pass



from PySide6.QtCore import QRect
from PySide6.QtGui import QColor, QIcon, QImage, QPixmap, Qt
from PySide6.QtWidgets import (
    QButtonGroup,
    QCheckBox,
    QComboBox,
    QFileDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QSpinBox,
    QVBoxLayout,
)

from foundry import data_dir, icon
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.objects.EnemyItem import MASK_COLOR
from foundry.gui.CustomDialog import CustomDialog
from foundry.gui.HorizontalLine import HorizontalLine
from foundry.gui.settings import (
    GUI_STYLE,
    RESIZE_LEFT_CLICK,
    RESIZE_RIGHT_CLICK,
    SETTINGS,
    load_settings,
    save_settings,
)
from foundry.smb3parse.constants import (
    POWERUP_FIREFLOWER,
    POWERUP_FROG,
    POWERUP_HAMMER,
    POWERUP_MUSHROOM,
    POWERUP_RACCOON,
    POWERUP_TANOOKI,
)

load_settings()

POWERUPS_NAME = 0
POWERUPS_X = 1
POWERUPS_Y = 2
POWERUPS_VALUE = 3
POWERUPS_PWING = 4
POWERUPS = [
    ("Small Mario", 32, 53, 0, False),
    ("Big Mario", 6, 48, POWERUP_MUSHROOM, False),
    ("Raccoon Mario", 57, 53, POWERUP_RACCOON, False),
    ("Fire Mario", 16, 53, POWERUP_FIREFLOWER, False),
    ("Tanooki Mario", 54, 53, POWERUP_TANOOKI, False),
    ("Frog Mario", 56, 53, POWERUP_FROG, False),
    ("Hammer Mario", 58, 53, POWERUP_HAMMER, False),
    # Even though P-Wing can *technically* be combined, it only really works with Raccoon and Tanooki suit
    ("Raccoon Mario with P-Wing", 55, 53, POWERUP_RACCOON, True),
    ("Tanooki Mario with P-Wing", 55, 53, POWERUP_TANOOKI, True),
]

png = QImage(str(data_dir / "gfx.png"))
png.convertTo(QImage.Format_RGB888)


class SettingsDialog(CustomDialog):
    def __init__(self, parent=None):
        super(SettingsDialog, self).__init__(parent, "Settings")

        mouse_box = QGroupBox("Mouse", self)
        mouse_box.setLayout(QVBoxLayout())

        label = QLabel("Scroll objects with mouse wheel:")
        label.setToolTip("Select an object and scroll up and down to change its type.")
        self._scroll_check_box = QCheckBox("Enabled")
        self._scroll_check_box.setChecked(SETTINGS["object_scroll_enabled"])
        self._scroll_check_box.toggled.connect(self._update_settings)

        scroll_layout = QHBoxLayout()
        scroll_layout.addWidget(label)
        scroll_layout.addStretch(1)
        scroll_layout.addWidget(self._scroll_check_box)

        label = QLabel("Show object names on hover:")
        label.setToolTip(
            "When hovering your cursor over an object in a level, its name and position is shown in a tooltip."
        )
        self._tooltip_check_box = QCheckBox("Enabled")
        self._tooltip_check_box.setChecked(SETTINGS["object_tooltip_enabled"])
        self._tooltip_check_box.toggled.connect(self._update_settings)

        tooltip_layout = QHBoxLayout()
        tooltip_layout.addWidget(label)
        tooltip_layout.addStretch(1)
        tooltip_layout.addWidget(self._tooltip_check_box)

        self.lmb_radio = QRadioButton("Left Mouse Button")
        rmb_radio = QRadioButton("Right Mouse Button")

        self.lmb_radio.setChecked(SETTINGS["resize_mode"] == RESIZE_LEFT_CLICK)
        rmb_radio.setChecked(SETTINGS["resize_mode"] == RESIZE_RIGHT_CLICK)

        self.lmb_radio.toggled.connect(self._update_settings)

        radio_group = QButtonGroup()
        radio_group.addButton(self.lmb_radio)
        radio_group.addButton(rmb_radio)

        resize_layout = QHBoxLayout()
        resize_layout.addWidget(QLabel("Object resize mode:"))
        resize_layout.addStretch(1)
        resize_layout.addWidget(self.lmb_radio)
        resize_layout.addWidget(rmb_radio)

        mouse_box.layout().addLayout(scroll_layout)
        mouse_box.layout().addLayout(tooltip_layout)
        mouse_box.layout().addLayout(resize_layout)

        # -----------------------------------------------
        # GUI theme section

        self.gui_style_box = QGroupBox("GUI", self)
        self.gui_style_hbox = QHBoxLayout(self.gui_style_box)
        self.gui_style_hbox.addWidget(QLabel("Style:"))
        self.gui_style_combo_box = QComboBox()
        self.gui_style_hbox.addWidget(self.gui_style_combo_box)

        for gui_style in GUI_STYLE.keys():
            self.gui_style_combo_box.addItem(gui_style.capitalize())
        self.gui_style_combo_box.setCurrentIndex(list(GUI_STYLE.keys()).index(SETTINGS["gui_style"]))
        self.gui_style_combo_box.currentIndexChanged.connect(self._update_settings)

        # -----------------------------------------------
        # emulator command

        self.emulator_command_input = QLineEdit(self)
        self.emulator_command_input.setPlaceholderText("Path to emulator")
        self.emulator_command_input.setText(SETTINGS["instaplay_emulator"])

        self.emulator_command_input.textChanged.connect(self._update_settings)

        self.emulator_path_button = QPushButton(icon("folder.svg"), "", self)
        self.emulator_path_button.pressed.connect(self._get_emulator_path)

        self.command_arguments_input = QLineEdit(self)
        self.command_arguments_input.setPlaceholderText("%f")
        self.command_arguments_input.setText(SETTINGS["instaplay_arguments"])

        self.command_arguments_input.textEdited.connect(self._update_settings)

        self.command_label = QLabel()

        command_box = QGroupBox("Emulator", self)
        command_layout = QVBoxLayout(command_box)

        command_layout.addWidget(QLabel('Emulator command or "path to exe":'))

        command_input_layout = QHBoxLayout()
        command_input_layout.addWidget(self.emulator_command_input)
        command_input_layout.addWidget(self.emulator_path_button)

        command_layout.addLayout(command_input_layout)
        command_layout.addWidget(QLabel("Command arguments (%f will be replaced with rom path):"))
        command_layout.addWidget(self.command_arguments_input)
        command_layout.addWidget(QLabel("Command used to play the rom:"))
        command_layout.addWidget(self.command_label)

        command_layout.addWidget(HorizontalLine())

        command_layout.addWidget(QLabel("Power up of Mario when playing level:"))
        self.powerup_combo_box = QComboBox()

        for name, x, y, value, p_wing in POWERUPS:
            powerup_icon = self._load_from_png(x, y)

            self.powerup_combo_box.addItem(powerup_icon, name)

        self.powerup_combo_box.currentIndexChanged.connect(self._update_settings)
        self.powerup_combo_box.setCurrentIndex(SETTINGS["default_powerup"])

        label = QLabel("Start with invincibility:")
        label.setToolTip("The player will start with star power.")
        self.powerup_star = QCheckBox("Enabled")
        self.powerup_star.setIcon(self._load_from_png(5, 48))
        self.powerup_star.setChecked(SETTINGS["default_power_has_star"])
        self.powerup_star.toggled.connect(self._update_settings)

        powerup_star_layout = QHBoxLayout()
        powerup_star_layout.addWidget(label)
        powerup_star_layout.addStretch(1)
        powerup_star_layout.addWidget(self.powerup_star)

        label = QLabel("Starting World:")
        label.setToolTip("The world that the player starts inside.")
        self.starting_world = QSpinBox(self)
        self.starting_world.setRange(1, 9)
        self.starting_world.setValue(SETTINGS["default_starting_world"])
        self.starting_world.valueChanged.connect(self._update_settings)

        starting_world_layout = QHBoxLayout()
        starting_world_layout.addWidget(label)
        starting_world_layout.addStretch(1)
        starting_world_layout.addWidget(self.starting_world)

        command_layout.addWidget(self.powerup_combo_box)
        command_layout.addLayout(powerup_star_layout)
        command_layout.addLayout(starting_world_layout)

        # ----------------------

        layout = QVBoxLayout(self)
        layout.addWidget(mouse_box)
        layout.addWidget(self.gui_style_box)
        layout.addWidget(command_box)

        self.update()

    def update(self):
        self.command_label.setText(f" > {SETTINGS['instaplay_emulator']} {SETTINGS['instaplay_arguments']}")

    def _update_settings(self, _):
        SETTINGS["instaplay_emulator"] = self.emulator_command_input.text()
        SETTINGS["instaplay_arguments"] = self.command_arguments_input.text()

        if self.lmb_radio.isChecked():
            SETTINGS["resize_mode"] = RESIZE_LEFT_CLICK
        else:
            SETTINGS["resize_mode"] = RESIZE_RIGHT_CLICK

        # setup style sheets
        SETTINGS["gui_style"] = self.gui_style_combo_box.currentText().upper()
        GUI_STYLE[SETTINGS["gui_style"]](self.parent())

        SETTINGS["object_scroll_enabled"] = self._scroll_check_box.isChecked()
        SETTINGS["object_tooltip_enabled"] = self._tooltip_check_box.isChecked()

        SETTINGS["default_powerup"] = self.powerup_combo_box.currentIndex()
        if hasattr(self, "powerup_star"):
            SETTINGS["default_power_has_star"] = self.powerup_star.isChecked()
        if hasattr(self, "starting_world"):
            SETTINGS["default_starting_world"] = self.starting_world.value()

        self.update()

    def _get_emulator_path(self):
        path_to_emulator, _ = QFileDialog.getOpenFileName(self, caption="Select emulator executable")

        if not path_to_emulator:
            return

        self.emulator_command_input.setText(path_to_emulator)

    @staticmethod
    def _load_from_png(x: int, y: int) -> QIcon:
        image = png.copy(QRect(x * Block.SIDE_LENGTH, y * Block.SIDE_LENGTH, Block.SIDE_LENGTH, Block.SIDE_LENGTH))
        mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
        image.setAlphaChannel(mask)

        pixmap = QPixmap.fromImage(image)
        icon_from_png = QIcon(pixmap)

        return icon_from_png

    def on_exit(self):
        save_settings()

        super(SettingsDialog, self).on_exit()



from PySide6.QtWidgets import QGridLayout, QGroupBox, QVBoxLayout

from foundry.game.gfx.Palette import (
    PALETTE_GROUPS_PER_OBJECT_SET,
    PALETTES_PER_PALETTES_GROUP,
    PaletteGroup,
)
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.CustomDialog import CustomDialog
from foundry.gui.PaletteWidget import PaletteWidget


class PaletteViewer(CustomDialog):
    palettes_per_row = 4

    def __init__(self, parent, level_ref: LevelRef):
        title = f"Palette Groups for Object Set {level_ref.level.object_set_number}"

        super(PaletteViewer, self).__init__(parent, title=title)

        self.level_ref = level_ref

        layout = QGridLayout(self)

        for palette_group in range(PALETTE_GROUPS_PER_OBJECT_SET):
            group_box = QGroupBox()
            group_box.setTitle(f"Palette Group {palette_group}")

            group_box_layout = QVBoxLayout(group_box)
            group_box_layout.setSpacing(0)

            pal = PaletteGroup.from_tileset(self.level_ref.level.object_set_number, palette_group)

            for idx in range(PALETTES_PER_PALETTES_GROUP):
                group_box_layout.addWidget(PaletteWidget(self, pal[idx]))

            row = palette_group // self.palettes_per_row
            col = palette_group % self.palettes_per_row

            layout.addWidget(group_box, row, col)



from functools import cache
from typing import Protocol

from attr import attrs
from PySide6.QtGui import QColor

from foundry import root_dir
from foundry.game.File import ROM
from foundry.smb3parse.constants import Palette_By_Tileset, PalSet_Maps
from foundry.smb3parse.levels import BASE_OFFSET

MAP_PALETTE_ADDRESS = PalSet_Maps

PRG_SIZE = 0x2000
PALETTE_PRG_NO = 22

PALETTE_BASE_ADDRESS = BASE_OFFSET + PALETTE_PRG_NO * PRG_SIZE
PALETTE_OFFSET_LIST = Palette_By_Tileset
PALETTE_OFFSET_SIZE = 2  # bytes

PALETTE_GROUPS_PER_OBJECT_SET = 8
ENEMY_PALETTE_GROUPS_PER_OBJECT_SET = 4
PALETTES_PER_PALETTES_GROUP = 4

COLORS_PER_PALETTE = 4
COLOR_SIZE = 1  # byte

PALETTE_DATA_SIZE = (
    (PALETTE_GROUPS_PER_OBJECT_SET + ENEMY_PALETTE_GROUPS_PER_OBJECT_SET)
    * PALETTES_PER_PALETTES_GROUP
    * COLORS_PER_PALETTE
)


class PaletteProtocol(Protocol):
    color_indexes: list[int]

    def __bytes__(self) -> bytes:
        ...

    def __getitem__(self, item: int) -> int:
        ...

    def __setitem__(self, key: int, value: int):
        ...

    @property
    def colors(self) -> list[QColor]:
        ...


class PaletteGroupProtocol(Protocol):
    palettes: list[PaletteProtocol]

    def __bytes__(self) -> bytes:
        ...

    def __getitem__(self, item: int) -> PaletteProtocol:
        ...

    def __setitem__(self, key: int, value: PaletteProtocol):
        ...

    @property
    def background_color(self) -> QColor:
        ...


@attrs(slots=True, auto_attribs=True)
class Palette:
    color_indexes: list[int]

    def __bytes__(self) -> bytes:
        return bytes([i & 0xFF for i in self.color_indexes])

    def __getitem__(self, item: int) -> int:
        return self.color_indexes[item]

    def __setitem__(self, key: int, value: int):
        self.color_indexes[key] = value

    @classmethod
    def as_empty(cls):
        """
        Makes an empty palette of default values

        Returns
        -------
        Palette
            A Palette filled with default values.
        """
        return cls([0, 0, 0, 0])

    @classmethod
    def from_rom(cls, address: int):
        """
        Creates a palette from an absolute address in ROM.

        Parameters
        ----------
        address : int
            The absolute address into the ROM.

        Returns
        -------
        Palette
            The Palette that represents the absolute address in ROM.
        """
        return cls([int(i) for i in ROM().read(address, COLORS_PER_PALETTE)])

    @property
    def colors(self) -> list[QColor]:
        return [NESPalette[c & 0x3F] for c in self.color_indexes]


@cache
def get_internal_palette_offset(tileset: int) -> int:
    """
    Provides the absolute internal position of the palette group offset from ROM.

    Parameters
    ----------
    tileset : int
        The tileset to find the absolute internal position of.

    Returns
    -------
    int
        The absolute internal position of the tileset's palette group.
    """
    return PALETTE_BASE_ADDRESS + ROM().little_endian(PALETTE_OFFSET_LIST + (tileset * PALETTE_OFFSET_SIZE))


@attrs(slots=True, auto_attribs=True)
class PaletteGroup:
    palettes: list[PaletteProtocol]

    def __bytes__(self) -> bytes:
        b = bytearray()
        for palette in self.palettes:
            b.extend(bytes(palette))
        return bytes(b)

    def __getitem__(self, item: int) -> PaletteProtocol:
        return self.palettes[item]

    def __setitem__(self, key: int, value: PaletteProtocol):
        self.palettes[key] = value

    @property
    def background_color(self) -> QColor:
        return self.palettes[0].colors[0]

    @classmethod
    def as_empty(cls):
        """
        Makes an empty palette group of default values

        Returns
        -------
        PaletteGroup
            A PaletteGroup filled with default values.
        """
        return cls([Palette.as_empty() for _ in range(PALETTES_PER_PALETTES_GROUP)])

    @classmethod
    def from_rom(cls, address: int):
        """
        Creates a palette group from an absolute address in ROM.

        Parameters
        ----------
        address : int
            The absolute address into the ROM.

        Returns
        -------
        PaletteGroup
            The PaletteGroup that represents the absolute address in ROM.
        """
        return cls(
            [
                Palette.from_rom(address + offset)
                for offset in [COLORS_PER_PALETTE * i for i in range(PALETTES_PER_PALETTES_GROUP)]
            ]
        )

    @classmethod
    def from_tileset(cls, tileset: int, index: int):
        """
        Loads a palette group from a tileset with a given index.

        Parameters
        ----------
        tileset : int
            The index of the tileset.
        index : int
            The index of the palette group inside the tileset.

        Returns
        -------
        PaletteGroup
            The PaletteGroup that represents the tileset's palette group at the provided offset.
        """
        offset = get_internal_palette_offset(tileset) + index * PALETTES_PER_PALETTES_GROUP * COLORS_PER_PALETTE
        return cls.from_rom(offset)


palette_file = root_dir.joinpath("data", "Default.pal")

with open(palette_file, "rb") as f:
    color_data = f.read()

offset = 0x18  # first color position

NESPalette: list[QColor] = []
COLOR_COUNT = 64
BYTES_IN_COLOR = 3 + 1  # bytes + separator

for i in range(COLOR_COUNT):
    NESPalette.append(QColor(color_data[offset], color_data[offset + 1], color_data[offset + 2]))

    offset += BYTES_IN_COLOR



from typing import Protocol

from PySide6.QtGui import QAction

from foundry.gui.ContextMenu import ContextMenu
from foundry.gui.JumpList import JumpList
from foundry.gui.LevelSizeBar import LevelSizeBar
from foundry.gui.LevelView import LevelView
from foundry.gui.ObjectDropdown import ObjectDropdown
from foundry.gui.ObjectList import ObjectList
from foundry.gui.ObjectToolBar import ObjectToolBar
from foundry.gui.PaletteGroupController import PaletteGroupController
from foundry.gui.SpinnerPanel import SpinnerPanel
from foundry.gui.WarningList import WarningList


class LevelControlled(Protocol):
    context_menu: ContextMenu
    level_view: LevelView
    spinner_panel: SpinnerPanel
    object_list: ObjectList
    object_dropdown: ObjectDropdown
    level_size_bar: LevelSizeBar
    enemy_size_bar: LevelSizeBar
    side_palette: PaletteGroupController
    jump_list: JumpList
    object_toolbar: ObjectToolBar
    warning_list: WarningList
    undo_action: QAction
    redo_action: QAction
    jump_destination_action: QAction
    menu_toolbar_save_action: QAction

    @property
    def safe_to_change(self) -> bool:
        ...

    def update_title(self):
        ...



from typing import Union

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QVBoxLayout, QWidget

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.ObjectLike import ObjectLike
from foundry.gui.TabbedToolBox import TabbedToolBox


class ObjectToolBar(QWidget):
    selected: SignalInstance = Signal(ObjectLike)  # type: ignore

    def __init__(self, parent=None):
        super(ObjectToolBar, self).__init__(parent)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)
        self.setMinimumHeight(200)
        self.setMinimumWidth(50)
        self.tileset = -1

        self.tool_box = TabbedToolBox()
        self.tool_box.selected_index.connect(self.select_object)

        layout.addWidget(self.tool_box, stretch=1)

    def set_object_set(self, object_set_index: int, graphic_set_index: int = -1):
        if self.tileset == object_set_index:
            return
        self.tileset = object_set_index
        self.tool_box.set_object_set(object_set_index, graphic_set_index)

    def select_object(self, tab_index: int, object_index: int):
        item = self.tool_box.select_object(tab_index, object_index)
        self.selected.emit(item)

    def add_recent_object(self, level_object: Union[EnemyObject, LevelObject]):
        self.tool_box.add_recent_object(level_object)



from typing import Optional

from PySide6.QtCore import QSize, Signal, SignalInstance
from PySide6.QtGui import QColor, QMouseEvent, QPixmap, Qt
from PySide6.QtWidgets import QLabel, QSizePolicy, QWidget


class ColorButtonWidget(QLabel):
    """
    A colored button that can be pressed and interacted with

    Attributes
    ----------
    clicked: Signal
        Slot associated with the button being clicked.
    size_changed: Signal[QSize]
        Slot associated with the button changing size.
    color_changed: Signal[QColor]
        Slot associated with the button changing color.
    selected_changed: Signal[bool]
        Slot associated with the button changing its selected state.
    """

    clicked: SignalInstance = Signal()  # type: ignore
    size_changed: SignalInstance = Signal(QSize)  # type: ignore
    color_changed: SignalInstance = Signal(QColor)  # type: ignore
    selected_changed: SignalInstance = Signal(bool)  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        color: Optional[QColor] = None,
        size: Optional[QSize] = None,
        selected: bool = False,
    ):
        super().__init__(parent)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        if color is not None:
            self._color = color
        else:
            self._color = QColor(Qt.white)
        if size is not None:
            self._size_ = size
        else:
            self._size_ = QSize(16, 16)
        self._selected = selected
        self._update()

    @property
    def size_(self) -> QSize:
        return self._size_

    @size_.setter
    def size_(self, size: QSize):
        self._size_ = size
        self.size_changed.emit(size)
        self._update()

    @property
    def color(self) -> QColor:
        return self._color

    @color.setter
    def color(self, color: QColor):
        self._color = color
        self.color_changed.emit(color)
        self._update()

    @property
    def selected(self) -> bool:
        return self._selected

    @selected.setter
    def selected(self, selected: bool):
        self._selected = selected
        self.selected_changed.emit(selected)
        self._update()

    def _update(self):
        pix = QPixmap(self.size_)
        pix.fill(self.color)
        self.setPixmap(pix)

        if self.selected:
            if self.color.lightnessF() < 0.25:
                self.setStyleSheet("border-color: rgb(255, 255, 255); border-width: 2px; border-style: solid")
            else:
                self.setStyleSheet("border-color: rgb(0, 0, 0); border-width: 2px; border-style: solid")
        else:
            rgb = self.color.getRgb()
            self.setStyleSheet(
                f"border-color: rgb({rgb[0]}, {rgb[1]}, {rgb[2]}); border-width: 2px; border-style: solid"
            )

        self.update()

    def mouseReleaseEvent(self, event: QMouseEvent):
        self.clicked.emit()



from typing import Protocol

from attr import attrs
from PySide6.QtCore import QPoint


class PositionProtocol(Protocol):
    """
    A two dimensional representation of a point on a plain.

    Attributes
    ----------
    x: int
        The horizontal location of the point.
    y: int
        The vertical location of the point.
    """

    x: int
    y: int


@attrs(slots=True, auto_attribs=True)
class Position:
    """
    A two dimensional representation of a point on a plain, that uses ``attrs`` to create a basic
    implementation.

    Attributes
    ----------
    x: int
        The horizontal location of the point.
    y: int
        The vertical location of the point.
    """

    x: int
    y: int

    @classmethod
    def from_qpoint(cls, point: QPoint):
        return cls(point.x(), point.y())



from enum import Enum
from functools import cache
from json import loads
from typing import Optional

from pydantic import BaseModel, Field

from foundry import enemy_definitions
from foundry.core.warnings.Warning import Warning
from foundry.game.Definitions import Definition


class GeneratorType(int, Enum):
    """
    The various different ways enemies can be represented inside the game.
    """

    SINGLE_SPRITE_OBJECT = 0
    SINGLE_BLOCK_OBJECT = 9
    CENTERED = 10


class Sprite(BaseModel):
    """
    Defines a given sprite and its properties
    """

    index: int
    palette_index: int = 1
    x_offset: int = 0
    y_offset: int = 0
    horizontal_mirror: bool = False
    vertical_mirror: bool = False


class EnemyDefinition(Definition):
    bmp_width: int = 1
    bmp_height: int = 1
    rect_width: int = 0
    rect_height: int = 0
    rect_x_offset: int = 0
    rect_y_offset: int = 0
    icon_width: int = 0
    icon_height: int = 0
    icon_x_offset: Optional[int] = None
    icon_y_offset: Optional[int] = None
    sprites: list[Sprite] = Field(default_factory=list)
    blocks: list[int] = Field(default_factory=list)
    pages: list[int] = Field(default_factory=list)
    orientation: GeneratorType = GeneratorType.SINGLE_BLOCK_OBJECT
    check_level_bounds: bool = True
    check_compatibility: bool = True

    @property
    def suggested_icon_width(self) -> int:
        """
        Provides the suggested icon width, as the icon width my not be set.

        Returns
        -------
        int
            The suggested icon width.
        """
        width = self.icon_width
        if not width:
            width = self.rect_width
            if not width:
                width = self.bmp_width
                if GeneratorType.SINGLE_SPRITE_OBJECT == self.orientation:
                    width //= 2
        return width

    @property
    def suggested_icon_height(self) -> int:
        """
        Provides the suggested icon height, as the icon height my not be set.

        Returns
        -------
        int
            The suggested icon height.
        """
        height = self.icon_height
        if not height:
            height = self.rect_height
            if not height:
                height = self.bmp_height
        return height

    @property
    def suggested_icon_x_offset(self) -> int:
        """
        Provides the suggested icon x offset, as the icon x offset may be None.

        Returns
        -------
        int
            The suggested icon x offset.
        """
        offset = self.icon_x_offset
        if offset is None:
            offset = self.rect_x_offset * 16
        return offset

    @property
    def suggested_icon_y_offset(self) -> int:
        """
        Provides the suggested icon y offset, as the icon y offset may be None.

        Returns
        -------
        int
            The suggested icon y offset.
        """
        offset = self.icon_y_offset
        if offset is None:
            offset = self.rect_y_offset * 16
        return offset

    def get_warnings(self) -> list[Warning]:
        warnings: list[Warning] = [warning.to_warning() for warning in self.warnings.copy()]  # type: ignore
        if self.check_compatibility:
            from foundry.core.warnings.EnemyCompatibilityWarning import (
                EnemyCompatibilityWarning,
            )

            warnings.append(EnemyCompatibilityWarning())
        if self.check_level_bounds:
            from foundry.core.warnings.OutsideLevelBoundsWarning import (
                OutsideLevelBoundsWarning,
            )

            warnings.append(OutsideLevelBoundsWarning())
        return warnings


class EnemyDefinitions(BaseModel):
    __root__: list[EnemyDefinition]


@cache
def get_enemy_metadata() -> EnemyDefinitions:
    with open(enemy_definitions, "r") as f:
        return EnemyDefinitions(__root__=loads(f.read()))



from attr import attrs
from pydantic import BaseModel

from foundry.core.warnings.WarningType import WarningType


@attrs(slots=True, frozen=True, auto_attribs=True)
class Warning:
    """
    A generic dataclass to represent a given warning.
    """

    def __init__(*args, **kwargs):
        pass

    def check_object(self, obj, *args, **kwargs) -> bool:
        """
        Determines if the object should emit a warning.

        Parameters
        ----------
        obj : ObjectLike
            The object to check.

        Returns
        -------
        bool
            If the object should emit a warning.
        """
        return True

    def get_message(self, obj, *args, **kwargs) -> str:
        """
        Provides a message for an object with a warning.

        Parameters
        ----------
        obj : ObjectLike
            The object to provide a warning for.

        Returns
        -------
        str
            The warning for the object.
        """
        return "Something is wrong"


class PydanticWarning(BaseModel):
    """
    A JSON model of a generic Warning through Pydantic

    Attributes
    ----------
    type: WarningType
        The type of warning.
    """

    type: WarningType

    class Config:
        use_enum_values = True

    def to_warning(self) -> Warning:
        """
        Converts self to a warning with respect to its type.

        Returns
        -------
        Warning
            That represents self.
        """
        from foundry.core.warnings.util import convert_pydantic_to_warning

        d = self.dict().copy()
        d.pop("type")  # The type is no longer needed
        return convert_pydantic_to_warning(self)(**d)
